import torch.nn as nn
import torch.nn.functional as func


class LeNet(nn.Module):
    def __init__(self):
        super(LeNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, kernel_size=5)
        self.conv2 = nn.Conv2d(6, 16, kernel_size=5)
        self.fc1 = nn.Linear(16*5*5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = func.relu(self.conv1(x))
        x = func.max_pool2d(x, 2)
        x = func.relu(self.conv2(x))
        x = func.max_pool2d(x, 2)
        x = x.view(x.size(0), -1)
        x = func.relu(self.fc1(x))
        x = func.relu(self.fc2(x))
        x = self.fc3(x)
        return x


import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import ConvModule


class FPNHead(nn.Module):
    """Panoptic Feature Pyramid Networks
    https://arxiv.org/abs/1901.02446
    """
    def __init__(self, in_channels, channel=128, num_classes=19):
        super().__init__()
        self.lateral_convs = nn.ModuleList([])
        self.output_convs = nn.ModuleList([])

        for ch in in_channels[::-1]:
            self.lateral_convs.append(ConvModule(ch, channel, 1))
            self.output_convs.append(ConvModule(channel, channel, 3, 1, 1))

        self.conv_seg = nn.Conv2d(channel, num_classes, 1)
        self.dropout = nn.Dropout2d(0.1)

    def forward(self, features) -> Tensor:
        features = features[::-1]
        out = self.lateral_convs[0](features[0])
        
        for i in range(1, len(features)):
            out = F.interpolate(out, scale_factor=2.0, mode='nearest')
            out = out + self.lateral_convs[i](features[i])
            out = self.output_convs[i](out)
        out = self.conv_seg(self.dropout(out))
        return out


if __name__ == '__main__':
    from semseg.models.backbones import ResNet
    backbone = ResNet('50')
    head = FPNHead([256, 512, 1024, 2048], 128, 19)
    x = torch.randn(2, 3, 224, 224)
    features = backbone(x)
    out = head(features)
    out = F.interpolate(out, size=x.shape[-2:], mode='bilinear', align_corners=False)
    print(out.shape)


import os
from typing import Optional, Union, Any

import pandas as pd
import numpy as np

import torch
import wandb
from PIL import Image
from matplotlib import pyplot as plt
from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.common.environment.ddp_utils import multi_process_safe

from ezdl.logger.basesg_logger import BaseSGLogger

logger = get_logger(__name__)

WANDB_ID_PREFIX = 'wandb_id.'
WANDB_INCLUDE_FILE_NAME = '.wandbinclude'


class WandBSGLogger(BaseSGLogger):

    def __init__(self, project_name: str, experiment_name: str, storage_location: str, resumed: bool,
                 training_params: dict, checkpoints_dir_path: str, tb_files_user_prompt: bool = False,
                 launch_tensorboard: bool = False, tensorboard_port: int = None, save_checkpoints_remote: bool = True,
                 save_tensorboard_remote: bool = True, save_logs_remote: bool = True, entity: Optional[str] = None,
                 api_server: Optional[str] = None, save_code: bool = False, tags=None, run_id=None, **kwargs):
        """

        :param experiment_name: Used for logging and loading purposes
        :param s3_path: If set to 's3' (i.e. s3://my-bucket) saves the Checkpoints in AWS S3 otherwise saves the Checkpoints Locally
        :param checkpoint_loaded: if true, then old tensorboard files will *not* be deleted when tb_files_user_prompt=True
        :param max_epochs: the number of epochs planned for this training
        :param tb_files_user_prompt: Asks user for Tensorboard deletion prompt.
        :param launch_tensorboard: Whether to launch a TensorBoard process.
        :param tensorboard_port: Specific port number for the tensorboard to use when launched (when set to None, some free port
                    number will be used
        :param save_checkpoints_remote: Saves checkpoints in s3.
        :param save_tensorboard_remote: Saves tensorboard in s3.
        :param save_logs_remote: Saves log files in s3.
        :param save_code: save current code to wandb
        """
        self.s3_location_available = storage_location.startswith('s3')
        self.resumed = resumed
        resume = 'must' if resumed else None
        os.makedirs(checkpoints_dir_path, exist_ok=True)
        self.run = wandb.init(project=project_name, name=experiment_name,
                         entity=entity, resume=resume, id=run_id, tags=tags,
                         dir=checkpoints_dir_path, **kwargs)
        if save_code:
            self._save_code()

        self.save_checkpoints_wandb = save_checkpoints_remote
        self.save_tensorboard_wandb = save_tensorboard_remote
        self.save_logs_wandb = save_logs_remote
        checkpoints_dir_path = os.path.relpath(self.run.dir, os.getcwd())
        super().__init__(project_name, experiment_name, storage_location, resumed, training_params,
                         checkpoints_dir_path, tb_files_user_prompt, launch_tensorboard, tensorboard_port,
                         self.s3_location_available, self.s3_location_available, self.s3_location_available)

        self._set_wandb_id(self.run.id)
        if api_server is not None:
            if api_server != os.getenv('WANDB_BASE_URL'):
                logger.warning(f'WANDB_BASE_URL environment parameter not set to {api_server}. Setting the parameter')
                os.putenv('WANDB_BASE_URL', api_server)

    @multi_process_safe
    def _save_code(self):
        """
        Save the current code to wandb.
        If a file named .wandbinclude is avilable in the root dir of the project the settings will be taken from the file.
        Otherwise, all python file in the current working dir (recursively) will be saved.
        File structure: a single relative path or a single type in each line.
        i.e:

        src
        tests
        examples
        *.py
        *.yaml

        The paths and types in the file are the paths and types to be included in code upload to wandb
        """
        base_path, paths, types = self._get_include_paths()

        if len(types) > 0:
            def func(path):
                for p in paths:
                    if path.startswith(p):
                        for t in types:
                            if path.endswith(t):
                                return True
                return False

            include_fn = func
        else:
            include_fn = lambda path: path.endswith(".py")

        if base_path != ".":
            wandb.run.log_code(base_path, include_fn=include_fn)
        else:
            wandb.run.log_code(".", include_fn=include_fn)


    @multi_process_safe
    def add_config(self, tag: str = None, config: dict = None):
        if tag:
            config = {tag: config}
        wandb.config.update(config, allow_val_change=self.resumed)

    @multi_process_safe
    def add_scalar(self, tag: str, scalar_value: float, global_step: int = 0):
        wandb.log(data={tag: scalar_value}, step=global_step)

    @multi_process_safe
    def add_scalars(self, tag_scalar_dict: dict, global_step: int = 0):
        wandb.log(data=tag_scalar_dict, step=global_step)

    @multi_process_safe
    def add_image(self, tag: str, image: Union[torch.Tensor, np.array, Image.Image], data_format='CHW', global_step: int = 0):
        if isinstance(image, torch.Tensor):
            image = image.cpu().detach().numpy()
        if image.shape[0] < 5:
            image = image.transpose([1, 2, 0])
        wandb.log(data={tag: wandb.Image(image, caption=tag)}, step=global_step)

    @multi_process_safe
    def add_images(self, tag: str, images: Union[torch.Tensor, np.array], data_format='NCHW', global_step: int = 0):

        wandb_images = []
        for im in images:
            if isinstance(im, torch.Tensor):
                im = im.cpu().detach().numpy()

            if im.shape[0] < 5:
                im = im.transpose([1, 2, 0])
            wandb_images.append(wandb.Image(im))
        wandb.log({tag: wandb_images}, step=global_step)

    @multi_process_safe
    def add_video(self, tag: str, video: Union[torch.Tensor, np.array], global_step: int = 0):

        if video.ndim > 4:
            for index, vid in enumerate(video):
                self.add_video(tag=f'{tag}_{index}', video=vid, global_step=global_step)
        else:
            if isinstance(video, torch.Tensor):
                video = video.cpu().detach().numpy()
            wandb.log({tag: wandb.Video(video, fps=4)}, step=global_step)

    @multi_process_safe
    def add_histogram(self, tag: str, values: Union[torch.Tensor, np.array], bins: str, global_step: int = 0):
        wandb.log({tag: wandb.Histogram(values, num_bins=bins)}, step=global_step)

    @multi_process_safe
    def add_plot(self, tag: str, values: pd.DataFrame, xtitle, ytitle, classes_marker):

        table = wandb.Table(columns=[classes_marker, xtitle, ytitle], dataframe=values)
        plt = wandb.plot_table(
            tag,
            table,
            {"x": xtitle, "y": ytitle, "class": classes_marker},
            {
                "title": tag,
                "x-axis-title": xtitle,
                "y-axis-title": ytitle,
            },
        )
        wandb.log({tag: plt})

    @multi_process_safe
    def add_text(self, tag: str, text_string: str, global_step: int = 0):
        wandb.log({tag: text_string}, step=global_step)

    @multi_process_safe
    def add_figure(self, tag: str, figure: plt.figure, global_step: int = 0):
        wandb.log({tag: figure}, step=global_step)

    @multi_process_safe
    def add_mask(self, tag: str, image, mask_dict, global_step: int = 0):
        wandb.log({tag: wandb.Image(image, masks=mask_dict)}, step=global_step)

    @multi_process_safe
    def add_table(self, tag, data, columns, rows):
        if isinstance(data, torch.Tensor):
            data = [[x.item() for x in row] for row in data]
        table = wandb.Table(data=data, rows=rows, columns=columns)
        wandb.log({tag: table})

    @multi_process_safe
    def close(self, really=False):
        if really:
            super().close()
            wandb.finish()

    @multi_process_safe
    def add_file(self, file_name: str = None):
        wandb.save(glob_str=os.path.join(self._local_dir, file_name), base_path=self._local_dir, policy='now')

    @multi_process_safe
    def add_summary(self, metrics: dict):
        wandb.summary.update(metrics)

    @multi_process_safe
    def upload(self):

        if self.save_tensorboard_wandb:
            wandb.save(glob_str=self._get_tensorboard_file_name(), base_path=self._local_dir, policy='now')

        if self.save_logs_wandb:
            wandb.save(glob_str=self.log_file_path, base_path=self._local_dir, policy='now')

    @multi_process_safe
    def add_checkpoint(self, tag: str, state_dict: dict, global_step: int = 0):
        name = f'ckpt_{global_step}.pth' if tag is None else tag
        if not name.endswith('.pth'):
            name += '.pth'

        path = os.path.join(self._local_dir, name)
        torch.save(state_dict, path)

        if self.save_checkpoints_wandb:
            if self.s3_location_available:
                self.model_checkpoints_data_interface.save_remote_checkpoints_file(self.experiment_name, self._local_dir, name)
            wandb.save(glob_str=path, base_path=self._local_dir, policy='now')

    def _get_tensorboard_file_name(self):
        try:
            tb_file_path = self.tensorboard_writer.file_writer.event_writer._file_name
        except RuntimeError as e:
            logger.warning('tensorboard file could not be located for ')
            return None

        return tb_file_path

    def _get_wandb_id(self):
        for file in os.listdir(self._local_dir):
            if file.startswith(WANDB_ID_PREFIX):
                return file.replace(WANDB_ID_PREFIX, '')

    def _set_wandb_id(self, id):
        for file in os.listdir(self._local_dir):
            if file.startswith(WANDB_ID_PREFIX):
                os.remove(os.path.join(self._local_dir, file))

    def add(self, tag: str, obj: Any, global_step: int = None):
        pass

    def _get_include_paths(self):
        """
        Look for .wandbinclude file in parent dirs and return the list of paths defined in the file.

        file structure is a single relative (i.e. src/) or a single type (i.e *.py)in each line.
        the paths and types in the file are the paths and types to be included in code upload to wandb
        :return: if file exists, return the list of paths and a list of types defined in the file
        """

        wandb_include_file_path = self._search_upwards_for_file(WANDB_INCLUDE_FILE_NAME)
        if wandb_include_file_path is not None:
            with open(wandb_include_file_path) as file:
                lines = file.readlines()

            base_path = os.path.dirname(wandb_include_file_path)
            paths = []
            types = []
            for line in lines:
                line = line.strip().strip('/n')
                if line == "" or line.startswith("#"):
                    continue

                if line.startswith('*.'):
                    types.append(line.replace('*', ''))
                else:
                    paths.append(os.path.join(base_path, line))
            return base_path, paths, types

        return ".", [], []

    @staticmethod
    def _search_upwards_for_file(file_name: str):
        """
        Search in the current directory and all directories above it for a file of a particular name.
        :param file_name: file name to look for.
        :return: pathlib.Path, the location of the first file found or None, if none was found
        """

        try:
            cur_dir = os.getcwd()
            while cur_dir != '/':
                if file_name in os.listdir(cur_dir):
                    return os.path.join(cur_dir, file_name)
                else:
                    cur_dir = os.path.dirname(cur_dir)
        except RuntimeError as e:
            return None

        return None

    def create_image_mask_sequence(self, name):
        self.sequences[name] = wandb.Table(["ID", "Image"])

    def add_image_mask_to_sequence(self, sequence_name, name, image, mask_dict):
        self.sequences[sequence_name].add_data(name, wandb.Image(image, masks=mask_dict))

    def add_image_mask_sequence(self, name):
        wandb.log({name: self.sequences[name]})

    def __repr__(self):
        return "WandbSGLogger"



import torch
import torchvision.datasets as datasets
from torch.utils.data import ConcatDataset
from torch.utils.data.distributed import DistributedSampler

from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.training import utils as core_utils
from super_gradients.training.datasets.mixup import CollateMixup
from super_gradients.training.datasets.samplers.repeated_augmentation_sampler import RepeatAugSampler
from super_gradients.training.exceptions.dataset_exceptions import IllegalDatasetParameterException
from super_gradients.training.utils import get_param
default_dataset_params = {"batch_size": 64, "val_batch_size": 200, "test_batch_size": 200, "dataset_dir": "./data/",
                          "s3_link": None}
LIBRARY_DATASETS = {
    "cifar10": {'class': datasets.CIFAR10, 'mean': (0.4914, 0.4822, 0.4465), 'std': (0.2023, 0.1994, 0.2010)},
    "cifar100": {'class': datasets.CIFAR100, 'mean': (0.5071, 0.4865, 0.4409), 'std': (0.2673, 0.2564, 0.2762)},
    "SVHN": {'class': datasets.SVHN, 'mean': None, 'std': None}
}

logger = get_logger(__name__)


class DatasetInterface:
    """
    DatasetInterface - This class manages all of the "communiation" the Model has with the Data Sets
    """

    def __init__(self, dataset_params={}, train_loader=None, val_loader=None, test_loader=None, classes=None):
        """
        @param train_loader: torch.utils.data.Dataloader (optional) dataloader for training.
        @param test_loader: torch.utils.data.Dataloader (optional) dataloader for testing.
        @param classes: list of classes.

        Note: the above parameters will be discarded in case dataset_params is passed.

        @param dataset_params:

            - `batch_size` : int (default=64)

                Number of examples per batch for training. Large batch sizes are recommended.

            - `val_batch_size` : int (default=200)

                Number of examples per batch for validation. Large batch sizes are recommended.

            - `dataset_dir` : str (default="./data/")

                Directory location for the data. Data will be downloaded to this directory when getting it from a
                remote url.

            - `s3_link` : str (default=None)

                remote s3 link to download the data (optional).

            - `aug_repeat_count` : int (default=0)

                amount of repetitions (each repetition of an example is augmented differently) for each
                 example for the trainset.

        """

        self.dataset_params = core_utils.HpmStruct(**default_dataset_params)
        self.dataset_params.override(**dataset_params)

        self.trainset, self.valset, self.testset = None, None, None
        self.train_loader, self.val_loader, self.test_loader = train_loader, val_loader, test_loader
        self.classes = classes
        self.batch_size_factor = 1

    def build_data_loaders(self, batch_size_factor=1, num_workers=8, train_batch_size=None, val_batch_size=None,
                           test_batch_size=None, distributed_sampler: bool = False):
        """

        define train, val (and optionally test) loaders. The method deals separately with distributed training and standard
        (non distributed, or parallel training). In the case of distributed training we need to rely on distributed
        samplers.
        :param batch_size_factor: int - factor to multiply the batch size (usually for multi gpu)
        :param num_workers: int - number of workers (parallel processes) for dataloaders
        :param train_batch_size: int - batch size for train loader, if None will be taken from dataset_params
        :param val_batch_size: int - batch size for val loader, if None will be taken from dataset_params
        :param distributed_sampler: boolean flag for distributed training mode
        :return: train_loader, val_loader, classes: list of classes
        """
        # CHANGE THE BATCH SIZE ACCORDING TO THE NUMBER OF DEVICES - ONLY IN NON-DISTRIBUTED TRAINING MODE
        # IN DISTRIBUTED MODE WE NEED DISTRIBUTED SAMPLERS
        # NO SHUFFLE IN DISTRIBUTED TRAINING

        aug_repeat_count = get_param(self.dataset_params, "aug_repeat_count", 0)
        if aug_repeat_count > 0 and not distributed_sampler:
            raise IllegalDatasetParameterException("repeated augmentation is only supported with DDP.")

        if distributed_sampler:
            self.batch_size_factor = 1
            train_sampler = RepeatAugSampler(self.trainset,
                                             num_repeats=aug_repeat_count) if aug_repeat_count > 0 else DistributedSampler(
                self.trainset)
            val_sampler = DistributedSampler(self.valset)
            test_sampler = DistributedSampler(self.testset) if self.testset is not None else None
            train_shuffle = False
        else:
            self.batch_size_factor = batch_size_factor
            train_sampler = None
            val_sampler = None
            test_sampler = None
            train_shuffle = True

        if train_batch_size is None:
            train_batch_size = self.dataset_params.batch_size * self.batch_size_factor
        if val_batch_size is None:
            val_batch_size = self.dataset_params.val_batch_size * self.batch_size_factor
        if test_batch_size is None:
            test_batch_size = self.dataset_params.test_batch_size * self.batch_size_factor

        train_loader_drop_last = core_utils.get_param(self.dataset_params, 'train_loader_drop_last', default_val=False)

        cutmix = core_utils.get_param(self.dataset_params, 'cutmix', False)
        cutmix_params = core_utils.get_param(self.dataset_params, 'cutmix_params')

        # WRAPPING collate_fn
        train_collate_fn = core_utils.get_param(self.trainset, 'collate_fn')
        val_collate_fn = core_utils.get_param(self.valset, 'collate_fn')
        test_collate_fn = core_utils.get_param(self.testset, 'collate_fn')

        if cutmix and train_collate_fn is not None:
            raise IllegalDatasetParameterException("cutmix and collate function cannot be used together")

        if cutmix:
            # FIXME - cutmix should be available only in classification dataset. once we make sure all classification
            # datasets inherit from the same super class, we should move cutmix code to that class
            logger.warning("Cutmix/mixup was enabled. This feature is currently supported only "
                           "for classification datasets.")
            train_collate_fn = CollateMixup(**cutmix_params)

        # FIXME - UNDERSTAND IF THE num_replicas VARIBALE IS NEEDED
        # train_sampler = DistributedSampler(self.trainset,
        #                                    num_replicas=distributed_gpus_num) if distributed_sampler else None
        # val_sampler = DistributedSampler(self.valset,
        #                                   num_replicas=distributed_gpus_num) if distributed_sampler else None

        self.train_loader = torch.utils.data.DataLoader(self.trainset,
                                                        batch_size=train_batch_size,
                                                        shuffle=train_shuffle,
                                                        num_workers=num_workers,
                                                        pin_memory=True,
                                                        sampler=train_sampler,
                                                        collate_fn=train_collate_fn,
                                                        drop_last=train_loader_drop_last)

        self.val_loader = torch.utils.data.DataLoader(self.valset,
                                                      batch_size=val_batch_size,
                                                      shuffle=False,
                                                      num_workers=num_workers,
                                                      pin_memory=True,
                                                      sampler=val_sampler,
                                                      collate_fn=val_collate_fn)

        if self.testset is not None:
            self.test_loader = torch.utils.data.DataLoader(self.testset,
                                                           batch_size=test_batch_size,
                                                           shuffle=False,
                                                           num_workers=num_workers,
                                                           pin_memory=True,
                                                           sampler=test_sampler,
                                                           collate_fn=test_collate_fn)

        self.classes = self.trainset.classes

    def get_data_loaders(self, **kwargs):
        """
        Get self.train_loader, self.val_loader, self.test_loader, self.classes.

        If the data loaders haven't been initialized yet, build them first.

        :param kwargs: kwargs are passed to build_data_loaders.

        """

        if self.train_loader is None and self.val_loader is None:
            self.build_data_loaders(**kwargs)

        return self.train_loader, self.val_loader, self.test_loader, self.classes

    def get_val_sample(self, num_samples=1):
        if num_samples > len(self.valset):
            raise Exception("Tried to load more samples than val-set size")
        if num_samples == 1:
            return self.valset[0]
        else:
            return self.valset[0:num_samples]

    def get_dataset_params(self):
        return self.dataset_params

    def print_dataset_details(self):
        logger.info("{} training samples, {} val samples, {} classes".format(len(self.trainset), len(self.valset),
                                                                             len(self.trainset.classes)))


import collections
import io
import os
from io import StringIO
from typing import Any, Mapping

import collections.abc
from ruamel.yaml import YAML

#
# def setup_mlflow(exp_name: str, description: str) -> str:
#     """
#     Setup mlflow tracking server and experiment
#     """
#     tracking_uri = registry_uri = "http://localhost:5000"
#     mlflow.set_tracking_uri(tracking_uri)
#     mlflow.set_registry_uri(registry_uri)
#     print('URI set!')
#     client = MlflowClient(tracking_uri, registry_uri)
#     exp_info = client.get_experiment_by_name(exp_name)
#     exp_id = exp_info.experiment_id if exp_info else \
#         client.create_experiment(exp_name, tags={'mlflow.note.content': description})
#     print('Experiment set')
#     return exp_id
#
#
# def mlflow_server(mlruns_folder=False):
#     """
#     Start mlflow server
#     """
#     cmd = ["mlflow", 'server']
#     if mlruns_folder:
#         cmd.extend(['--backend-store-uri', f'file:{mlruns_folder}'])
#     cmd_env = cmd_env = os.environ.copy()
#     child = subprocess.Popen(
#         cmd, env=cmd_env, universal_newlines=True, stdin=subprocess.PIPE
#     )
#     return child
#
#
# class MLRun(MlflowClient):
#     def __init__(self, exp_name: str, description: str, run_id: Optional[str] = None):
#         exp_id = setup_mlflow(exp_name, description)
#         super().__init__(mlflow.get_tracking_uri(), mlflow.get_registry_uri())
#         if run_id is None:
#             self.run = self.create_run(experiment_id=exp_id)
#         else:
#             self.run = self.get_run(run_id)
#
#     def log_params(self, params: Mapping):
#         params = mlflow_linearize(params)
#         for k, v in params.items():
#             self.log_param(k, v)
#
#     def log_metrics(self, metrics: Mapping):
#         for k, v in metrics.items():
#             self.log_metric(k, v)
#
#     def log_param(self, key: str, value: Any, run_id: str = None) -> None:
#         if run_id is None:
#             run_id = self.run.info.run_id
#         super().log_param(run_id, key, value)
#
#     def log_metric(self, key: str, value: Any, run_id: str = None,
#                    timestamp: Optional[int] = None,
#                    step: Optional[int] = None) -> None:
#
#         value = value.item() if type(value) == torch.Tensor else value
#
#         if run_id is None:
#             run_id = self.run.info.run_id
#         super().log_metric(run_id, key, value)


def mlflow_linearize(dictionary: Mapping) -> Mapping:
    """
    Linearize a nested dictionary concatenating keys in order to allow mlflow parameters recording.

    :param dictionary: nested dict
    :return: one level dict
    """
    exps = {}
    for key, value in dictionary.items():
        if isinstance(value, collections.abc.Mapping):
            exps = {**exps,
                    **{key + '.' + lin_key: lin_value for lin_key, lin_value in mlflow_linearize(value).items()}}
        else:
            exps[key] = value
    return exps


def nested_dict_update(d, u):
    if u is not None:
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = nested_dict_update(d.get(k, {}), v)
            else:
                d[k] = v
    return d


def dict_to_yaml_string(mapping: Mapping) -> str:
    """
    Convert a nested dictionary or list to a string
    """
    string_stream = StringIO()
    yaml = YAML()
    yaml.dump(mapping, string_stream)
    output_str = string_stream.getvalue()
    string_stream.close()
    return output_str


def yaml_string_to_dict(s):
    return YAML(typ='safe', pure=True).load(s)


def load_yaml(path, return_string=False):
    if hasattr(path, "readlines"):
        d = YAML(typ='safe', pure=True).load(path)
        if return_string:
            path.seek(0)
            return d, path.read().decode('utf-8')
    with open(path, 'r') as param_stream:
        d = YAML(typ='safe', pure=True).load(param_stream)
        if return_string:
            param_stream.seek(0)
            return d, str(param_stream.read())
    return d


def dict_to_yaml(d: Mapping):
    stream = io.StringIO()
    YAML().dump(d, stream)
    stream.seek(0)
    return stream.read()


def values_to_number(collec) -> Any:
    """
    Convert all values in a dictionary or list to numbers
    """
    if isinstance(collec, collections.abc.Mapping):
        for key, value in collec.items():
            collec[key] = values_to_number(value)
    elif isinstance(collec, list):
        return [values_to_number(v) for v in collec]
    elif isinstance(collec, str) and collec.startswith('00'):
        return collec
    elif isinstance(collec, int):
        return collec
    else:
        try:
            return float(collec)
        except (ValueError, TypeError):
            pass
    return collec


def filter_none(collec) -> Any:
    """
    Filter out None values from a dictionary or list
    """
    if isinstance(collec, collections.abc.Mapping):
        for key in list(collec.keys()):
            value = filter_none(collec[key])
            if value is None:
                del collec[key]
            else:
                collec[key] = value
        return collec
    elif isinstance(collec, list):
        return [filter_none(v) for v in collec if v is not None]
    else:
        return collec


def update_collection(collec, value, key=None):
    if isinstance(collec, dict):
        if isinstance(value, dict):
            for keyv, valuev in value.items():
                collec = update_collection(collec, valuev, keyv)
        elif key is not None:
            if value is not None:
                collec[key] = value
        else:
            collec = {**collec, **value} \
                if value is not None else collec
    else:
        collec = value if value is not None else collec
    return collec


def safe_execute(default, exception, function, *args):
    try:
        return function(*args)
    except exception:
        return default


def get_module_class_from_path(path):
    path = os.path.normpath(path)
    splitted = path.split(os.sep)
    module = ".".join(splitted[:-1])
    cls = splitted[-1]
    return module, cls


def recursive_get(dictionary, *keys):
    """
    Get a value from nested dicts passing the keys
    :param dictionary: The nested dicts
    :param keys: sequence of keys
    :return: the value
    """
    if len(keys) == 1:
        return dictionary.get(keys[0])
    else:
        if dictionary.get(keys[0]):
            return recursive_get(dictionary[keys[0]], *keys[1:])




import torch
import torch.nn as nn
from super_gradients.training.models import SgModule
from super_gradients.training.models.segmentation_models.regseg import RegSeg, DEFAULT_REGSEG48_BACKBONE_PARAMS, \
    DEFAULT_REGSEG48_DECODER_PARAMS, RegSegBackbone, RegSegDecoder, RegSegHead, DEFAULT_REGSEG48_HEAD_PARAMS
from super_gradients.training.utils import HpmStruct, get_param
from super_gradients.modules import ConvBNReLU


class RegSeg48(RegSeg):

    def __init__(self, arch_params: HpmStruct):
        num_classes = get_param(arch_params, "num_classes")
        input_channels = get_param(arch_params, "input_channels", 3)
        stem = ConvBNReLU(in_channels=input_channels, out_channels=32, kernel_size=3, stride=2, padding=1)
        backbone = RegSegBackbone(in_channels=32, backbone_config=DEFAULT_REGSEG48_BACKBONE_PARAMS)
        decoder = RegSegDecoder(
            backbone.get_backbone_output_number_of_channels(),
            DEFAULT_REGSEG48_DECODER_PARAMS
        )
        head = RegSegHead(decoder.out_channels, num_classes, DEFAULT_REGSEG48_HEAD_PARAMS)
        super().__init__(stem, backbone, decoder, head)

    def replace_head(self, new_num_classes: int, head_config: dict = None):
        head_config = head_config or DEFAULT_REGSEG48_HEAD_PARAMS
        super().replace_head(new_num_classes, head_config)



from ezdl.models.regseg import RegSeg48
from ezdl.models.backbones import ResNet
from ezdl.models.segnet import SegNet
from ezdl.models.dummy import Random, SingleConv
from ezdl.models.deeplab import deeplabv3_resnet50

MODELS = {
    'segnet': SegNet,
    'regseg48': RegSeg48,
    'random': Random,
    'conv': SingleConv,
    'resnet': ResNet,
    'deeplabv3_resnet50': deeplabv3_resnet50
}



import os
import numbers
from typing import Iterable

import PIL
import torch
from torch import Tensor
from torch.nn.functional import one_hot

from torchvision.transforms import functional as F, InterpolationMode

from PIL import ImageOps


class PairSystematicTransform:

    def __init__(self, init_state, periodicity=1):
        self.periodicity = periodicity
        self.init_state = init_state

        self.pair_state = {}
        self.period_state = {}

    def __call__(self, img):
        pid = os.getpid()
        if pid in self.pair_state:
            state = self.pair_state.pop(pid)
        else:
            state, counter = self.period_state.get(pid) or (self.init_state, 0)
            self.pair_state[pid] = state
            if counter + 1 >= self.periodicity:
                self.period_state[pid] = self.change_state(state), 0
            else:
                self.period_state[pid] = state, counter + 1
        return self.transform(img, state)

    def transform(self, img, state):
        raise NotImplementedError

    def change_state(self, state):
        raise NotImplementedError


class PairFlip(PairSystematicTransform):
    def __init__(self, periodicity=1, orientation='horizontal'):
        super().__init__(False, periodicity)
        if orientation == 'horizontal':
            self.flip = F.hflip
        elif orientation == 'vertical':
            self.flip = F.vflip
        else:
            raise ValueError(f'Unknown orientation: {orientation}')

    def transform(self, img, state):
        if state:
            return self.flip(img)
        else:
            return img

    def change_state(self, state):
        return not state


class PairFourCrop(PairSystematicTransform):
    def __init__(self, size, padding=0, periodicity=1):
        super().__init__(0, periodicity)
        self.padding = padding
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def transform(self, img, state):
        if isinstance(img, PIL.Image.Image):
            h, w = img.size
        elif isinstance(img, torch.Tensor):
            w, h = img.size()[-2:]
        else:
            raise TypeError("img should be PIL.Image or torch.Tensor, not {}".format(type(img)))
        if state == 0:
            return F.crop(img, 0, 0, *self.size)
        elif state == 1:
            return F.crop(img, 0, h - self.size[1], *self.size)
        elif state == 2:
            return F.crop(img, w - self.size[0], 0, *self.size)
        elif state == 3:
            return F.crop(img, w - self.size[0], h - self.size[1], *self.size)

    def change_state(self, state):
        return (state + 1) % 4


class PairRandomCrop:
    """Crop the given PIL.Image at a random location.
    ** This is a MODIFIED version **, which supports identical random crop for
    both image and target map in Semantic Segmentation.
    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        padding (int or sequence, optional): Optional padding on each border
            of the image. Default is 0, i.e no padding. If a sequence of length
            4 is provided, it is used to pad left, top, right, bottom borders
            respectively.
    """
    image_crop_position = {}

    def __init__(self, size, padding=0):

        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding

    def __call__(self, img):
        """
        Args:
            img (PIL.Image): Image to be cropped.
        Returns:
            PIL.Image: Cropped image.
        """
        if self.padding > 0:
            img = ImageOps.expand(img, border=self.padding, fill=0)

        if isinstance(img, PIL.Image.Image):
            w, h = img.size
        elif isinstance(img, torch.Tensor):
            w, h = img.size()[-2:]
        else:
            raise TypeError("img should be PIL.Image or torch.Tensor, not {}".format(type(img)))

        th, tw = self.size
        if w == tw and h == th:
            return img

        pid = os.getpid()
        if pid in self.image_crop_position:
            x1, y1 = self.image_crop_position.pop(pid)
        else:
            x1 = torch.randint(size=(1,), low=0, high=w - tw)
            y1 = torch.randint(size=(1,), low=0, high=w - th)
            self.image_crop_position[pid] = (x1, y1)
        return F.crop(img, x1, y1, tw, th)


class PairRandomFlip(torch.nn.Module):
    """Flip the given image randomly with a given probability.
    If the image is torch Tensor, it is expected
    to have [..., H, W] shape, where ... means an arbitrary number of leading
    dimensions

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5, orientation='horizontal'):
        super().__init__()
        # _log_api_usage_once(self)
        self.p = p
        if orientation == 'horizontal':
            self.flip = F.hflip
        elif orientation == 'vertical':
            self.flip = F.vflip
        else:
            raise ValueError(f'Unknown orientation: {orientation}')
        self.image_flip = {}

    def forward(self, img):
        """
        Args:
            img (PIL Image or Tensor): Image to be flipped.

        Returns:
            PIL Image or Tensor: Randomly flipped image.
        """
        pid = os.getpid()
        if pid in self.image_flip:
            value = self.image_flip.pop(pid)
        else:
            value = torch.rand(1)
            self.image_flip[pid] = value
        if value < self.p:
            return self.flip(img)
        return img

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(p={self.p})"


class PairRandomRotation(torch.nn.Module):
    """Rotate the given image randomly with a given rotation.
    If the image is torch Tensor, it is expected
    to have [..., H, W] shape, where ... means an arbitrary number of leading
    dimensions

    Args:
        degree: degree of rotation.
        interpolation: Default: InterpolationMode.BILINEAR
    """

    def __init__(self, degree, interpolation=InterpolationMode.BILINEAR):
        super().__init__()
        # _log_api_usage_once(self)

        if isinstance(self.degree, Iterable):
            self.degree = degree
        else:
            self.degree = [-degree, degree]
        self.interpolation = interpolation
        self.image_rotation = {}

    def forward(self, img):
        """
        Args:
            img (PIL Image or Tensor): Image to be rotated.

        Returns:
            PIL Image or Tensor: Randomly rotated image.
        """
        pid = os.getpid()
        if pid in self.image_rotation:
            value = self.image_rotation.pop(pid)
        else:
            value = torch.FloatTensor(1).uniform_(*self.degree)
            self.image_rotation[pid] = value

        return F.rotate(img, value, self.interpolation)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(p={self.degree})"


class ToLong:
    def __call__(self, x):
        return x.long()


class SegOneHot:
    def __init__(self, num_classes):
        self.num_classes = num_classes

    def __call__(self, x):
        y = torch.moveaxis(one_hot(x, self.num_classes), 2, 0)
        return y


class FixValue:
    def __init__(self, source, target):
        self.s = source
        self.t = target

    def __call__(self, x):
        x[x == self.s] = self.t
        return x


class Denormalize(torch.nn.Module):
    """Denormalize a tensor image with mean and standard deviation.
    This transform does not support PIL Image.
    Given mean: ``(mean[1],...,mean[n])`` and std: ``(std[1],..,std[n])`` for ``n``
    channels, this transform will normalize each channel of the input
    ``torch.*Tensor`` i.e.,
    ``output[channel] = (input[channel] - mean[channel]) / std[channel]``

    .. note::
        This transform acts out of place, i.e., it does not mutate the input tensor.

    Args:
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
        inplace(bool,optional): Bool to make this operation in-place.

    """

    def __init__(self, mean, std, inplace=False):
        super().__init__()
        # _log_api_usage_once(self)
        self.mean = mean
        self.std = std
        self.inplace = inplace

    def forward(self, tensor: Tensor) -> Tensor:
        """
        Args:
            tensor (Tensor): Tensor image to be normalized.

        Returns:
            Tensor: Normalized Tensor image.
        """
        mean = torch.as_tensor(self.mean, dtype=tensor.dtype, device=tensor.device)
        std = torch.as_tensor(self.std, dtype=tensor.dtype, device=tensor.device)
        return tensor.mul(std.view(-1, 1, 1)).add(mean.view(-1, 1, 1))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(mean={self.mean}, std={self.std})"


def squeeze0(x):
    return torch.squeeze(x, dim=0)



import os.path

import torch
from PIL import Image
from super_gradients.training import StrictLoad
from super_gradients.training.utils.checkpoint_utils import load_checkpoint_to_model
from super_gradients.training import utils as core_utils
from torchvision.transforms import functional as F
import torchvision.transforms as transforms

from ezdl.utils.utilities import load_yaml

from ezdl.models import MODELS as MODELS_DICT


class Inferencer:
    def __init__(self, name, path, params, gpu=True):
        self.preprocess = None

        if name in MODELS_DICT.keys():
            self.model = MODELS_DICT[name](params)
        else:
            self.model = name
        self.model = core_utils.WrappedModel(self.model)
        _ = load_checkpoint_to_model(ckpt_local_path=path,
                                     load_backbone=False,
                                     net=self.model,
                                     strict=StrictLoad.ON.value,
                                     load_weights_only=True)
        self.model.eval()
        if gpu:
            self.model.cuda()

    def set_preprocess(self, preprocess_fn):
        self.preprocess = preprocess_fn

    def __call__(self, image, preprocess=True):
        if preprocess:
            image = self.preprocess(image)
        if next(self.model.parameters()).is_cuda:
            image = image.cuda()
        return self.model(image)

    def cuda(self):
        self.model.cuda()

    def cpu(self):
        self.model.cpu()


class WandbInferencer(Inferencer):
    config_file = 'config.yaml'
    config_key = 'value'
    wb_input_params = 'in_params'

    def __init__(self, path_wrapper, config_wrapper, gpu=True):
        self.channels = None
        config = load_yaml(config_wrapper.name)[self.wb_input_params]
        if self.config_key in config:
            config = config[self.config_key]

        input_channels = len(config['dataset']['channels'])
        output_channels = config['dataset']['num_classes']
        params = {
            'input_channels': input_channels,
            'output_channels': output_channels,
            'in_channels': input_channels,
            'out_channels': output_channels,
            'num_classes': output_channels,
            **config['model']['params']
        }
        super().__init__(name=config['model']['name'], path=path_wrapper.name, params=params, gpu=gpu)
        self.compose_preprocess(config)

    def compose_preprocess(self, config):  # TO DO
        pass
        # means, stds = WeedMapDatasetInterface.get_mean_std(
        #     config["dataset"]["train_folders"],
        #     config["dataset"]["channels"],
        #     os.path.basename(config["dataset"]["root"].lower()))
        # self.channels = config["dataset"]["channels"]
        # self.preprocess = transforms.Normalize(means, stds)


def compose_images(images: list):
    if len(images) == 1:
        return F.to_tensor(images[0]).unsqueeze(0)
    else:
        return torch.stack([F.to_tensor(img).squeeze(0) for img in images]).unsqueeze(0)



import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import ConvModule


class FCNHead(nn.Module):
    def __init__(self, c1, c2, num_classes: int = 19):
        super().__init__()
        self.conv = ConvModule(c1, c2, 1)
        self.cls = nn.Conv2d(c2, num_classes, 1)

    def forward(self, features) -> Tensor:
        x = self.conv(features[-1])
        x = self.cls(x)
        return x


if __name__ == '__main__':
    from semseg.models.backbones import ResNet
    backbone = ResNet('50')
    head = FCNHead(2048, 256, 19)
    x = torch.randn(2, 3, 224, 224)
    features = backbone(x)
    out = head(features)
    out = F.interpolate(out, size=x.shape[-2:], mode='bilinear', align_corners=False)
    print(out.shape)



from __future__ import annotations

import copy
import gc
import os
from typing import Mapping
from easydict import EasyDict

from ezdl.experiment.resume import ExpLog
from ezdl.experiment.run import Run
from ezdl.experiment.resume import get_interrupted_run, retrieve_run_to_resume
from ezdl.utils.grid import make_grid, linearize
from ezdl.utils.utilities import nested_dict_update, update_collection
from super_gradients.common.abstractions.abstract_logger import get_logger

logger = get_logger(__name__)


class GridSummary:
    def __init__(self,
                 total_runs,
                 total_runs_excl_grid,
                 total_runs_to_run,
                 total_runs_excl,
                 ):
        self.total_runs = total_runs
        self.total_runs_excl_grid = total_runs_excl_grid
        self.total_runs_to_run = total_runs_to_run
        self.total_runs_excl = total_runs_excl

    def update(self, d):
        self.total_runs = d.get("total_runs") or self.total_runs
        self.total_runs_excl_grid = d.get("total_runs_excl_grid") or self.total_runs_excl_grid
        self.total_runs_to_run = d.get("total_runs_to_run") or self.total_runs_to_run
        self.total_runs_excl = d.get("total_runs_excl") or self.total_runs_to_run


class ExpSettings(EasyDict):
    def __init__(self, *args, **kwargs):
        self.start_from_grid = 0
        self.start_from_run = 0
        self.resume = False
        self.resume_last = False
        self.tracking_dir = ""
        self.excluded_files = ""
        self.name = ""
        self.group = ""
        self.continue_with_errors = True
        self.logger = None
        super().__init__(*args, **kwargs)
        self.tracking_dir = self.tracking_dir or ""

    def update(self, e: ExpSettings, **f):
        if e is None:
            return
        self.start_from_grid = e.start_from_grid or self.start_from_grid
        self.start_from_run = e.start_from_run or self.start_from_run
        self.resume = e.resume or self.resume
        self.resume_last = e.resume_last or self.resume_last
        self.tracking_dir = e.tracking_dir or self.tracking_dir
        self.excluded_files = e.excluded_files or self.excluded_files
        self.group = e.group or self.group
        self.logger = e.logger or self.logger
        self.continue_with_errors = not e.continue_with_errors or self.continue_with_errors


class Status:
    STARTING = "starting"
    CRASHED = "crashed"
    FINISHED = "finished"

    def __init__(self, grid, run, params, n_grids, grid_len, wandb_run=None):
        self.grid = grid
        self.run = run
        self.params = params
        self.status = self.STARTING
        self.grid_len = grid_len
        self.n_grids = n_grids
        self.exception = None
        self.wandb_run = wandb_run

    def finish(self):
        self.params = {}
        self.status = self.FINISHED
        return self

    def crash(self, exception):
        self.status = self.CRASHED
        self.exception = exception
        return self


class StatusManager:
    def __init__(self, n_grids):
        self.n_grids = n_grids
        self.cur_status = None

    def new_run(self, grid, run, params, grid_len, wandb_run=None):
        self.cur_status = Status(grid=grid, run=run, params=params,
                                 n_grids=self.n_grids, grid_len=grid_len, wandb_run=wandb_run)
        return self.cur_status

    def update_run(self, wandb_run):
        self.cur_status.wandb_run = wandb_run
        return self.cur_status

    def finish_run(self):
        return self.cur_status.finish()

    def crash_run(self, exception):
        return self.cur_status.crash_run(exception)


class Experimenter:
    def __init__(self):
        self.gs = None
        self.exp_settings = ExpSettings()
        self.grids = None

    def calculate_runs(self, settings):
        base_grid = settings['parameters']
        other_grids = settings['other_grids']
        self.exp_settings = ExpSettings(settings['experiment'])

        complete_grids = [base_grid]
        if other_grids:
            complete_grids += \
                [nested_dict_update(copy.deepcopy(base_grid), other_run) for other_run in other_grids]
        logger.info(f'There are {len(complete_grids)} grids')

        self.grids, dot_elements = zip(*[make_grid(grid, return_cartesian_elements=True) for grid in complete_grids])
        # WARNING: Grids' objects have the same IDs!
        dot_elements = list(dot_elements)
        if len(dot_elements) > 1:
            dot_elements[1:] = [list(dict(linearize(others) + dot).items()) for others, dot in
                                zip(other_grids, dot_elements[1:])]

        # Modify starting grid and run to manage the resume
        self.manage_resume()

        for i, grid in enumerate(self.grids):
            info = f'Found {len(grid)} runs from grid {i}'
            if i < self.exp_settings.start_from_grid:
                info += f', skipping grid {i} with {len(grid)} runs'
            logger.info(info)
        self.generate_grid_summary()

        if self.exp_settings.excluded_files:
            os.environ['WANDB_IGNORE_GLOBS'] = self.exp_settings.excluded_files
        return self.gs, self.grids, dot_elements

    def generate_grid_summary(self):
        total_runs = sum(len(grid) for grid in self.grids)
        total_runs_excl_grid = total_runs - sum(
            len(grid) for grid in self.grids[self.exp_settings.start_from_grid :]
        )
        total_runs_excl = total_runs_excl_grid + self.exp_settings.start_from_run
        total_runs_to_run = total_runs - total_runs_excl
        self.gs = GridSummary(
            total_runs=total_runs,
            total_runs_excl_grid=total_runs_excl_grid,
            total_runs_to_run=total_runs_to_run,
            total_runs_excl=total_runs_excl
        )

    def execute_runs_generator(self):
        track_dir = self.exp_settings['tracking_dir']
        if track_dir:
            os.makedirs(track_dir, exist_ok=True)
        exp_log = ExpLog(track_dir, self.exp_settings.name, self.exp_settings.group)
        starting_run = self.exp_settings.start_from_run
        status_manager = StatusManager(len(self.grids))
        if self.exp_settings.resume_last:
            logger.info("+ another run to finish!")
            grid_len = len(self.grids[self.exp_settings.start_from_grid])
            grid_list = [(i, j) for i in range(len(self.grids)) for j in range(len(self.grids[i]))]
            index = grid_list.index((self.exp_settings.start_from_grid, self.exp_settings.start_from_run))
            sg, sr = grid_list[index - 1]
            try:
                exp_log.insert_run(sg, sr)
                run = get_interrupted_run(self.exp_settings)
                yield status_manager.new_run(sg, sr, run.params, grid_len, run)
                logger.info(f'Running grid {sg} out of {len(self.grids) - 1}')
                logger.info(
                    f'Running run {sr - 1} out of {grid_len} ({sum(len(self.grids[k]) for k in range(sg)) + sr} / {self.gs.total_runs - 1})'
                )
                run.launch()
                exp_log.finish_run(sg, sr)
                yield status_manager.finish_run()
            except Exception as e:
                logger.error(f'Experiment {sg} failed with error {e}')
                exp_log.finish_run(sg, sr, crashed=True)
                if not self.exp_settings.continue_with_errors:
                    raise e
                yield status_manager.crash_run(e)
        for i in range(self.exp_settings.start_from_grid, len(self.grids)):
            grid = self.grids[i]
            if i != self.exp_settings.start_from_grid:
                starting_run = 0
            for j in range(starting_run, len(grid)):
                params = grid[j]
                try:
                    exp_log.insert_run(i, j)
                    yield status_manager.new_run(i, j, params, len(grid))
                    logger.info(f'Running grid {i} out of {len(self.grids) - 1}')
                    logger.info(
                        f'Running run {j} out of {len(grid) - 1} ({sum(len(self.grids[k]) for k in range(i)) + j} / {self.gs.total_runs - 1})'
                    )
                    run = Run()
                    run.init({'experiment': {**self.exp_settings}, **params})
                    yield status_manager.update_run(run.seg_trainer.sg_logger.run)
                    run.launch()
                    exp_log.finish_run(i, j)
                    gc.collect()
                    yield status_manager.finish_run()
                except Exception as e:
                    logger.error(f'Experiment {i} failed with error {e}')
                    exp_log.finish_run(i, j, crashed=True)
                    if not self.exp_settings.continue_with_errors:
                        raise e
                    yield status_manager.crash_run(e)

    def execute_runs(self):
        for _ in self.execute_runs_generator():
            pass

    def manage_resume(self):
        if self.exp_settings.resume:
            self.exp_settings.start_from_grid, \
                self.exp_settings.start_from_run, \
                self.exp_settings.resume_last = retrieve_run_to_resume(self.exp_settings, self.grids)
        else:
            self.exp_settings.resume_last = False

    def update_settings(self, d):
        self.exp_settings = update_collection(self.exp_settings, d)
        if self.gs is None:
            return
        self.gs.update(self.exp_settings)
        if "resume" in d:
            self.manage_resume()
            self.generate_grid_summary()


def experiment(settings: Mapping, param_path: str = "local variable"):
    logger.info(f'Loaded parameters from {param_path}')

    experimenter = Experimenter()
    grid_summary, grids, cartesian_elements = experimenter.calculate_runs(settings)

    logger.info(f'Total runs found:              {grid_summary.total_runs}')
    logger.info(f'Total runs excluded by grids:  {grid_summary.total_runs_excl_grid}')
    logger.info(f'Total runs excluded:           {grid_summary.total_runs_excl}')
    logger.info(f'Total runs to run:             {grid_summary.total_runs_to_run}')
    experimenter.execute_runs()



import argparse
from ezdl.app.gui import streamlit_entry

parser = argparse.ArgumentParser(description='EzDL App')

parser.add_argument('--resume', required=False, action='store_true',
                    help='Resume the experiment', default=False)
parser.add_argument('-d', '--dir', required=False, type=str,
                    help='Set the local tracking directory', default=None)
parser.add_argument('-f', '--file', required=False, type=str,
                    help='Set the config file', default=None)
parser.add_argument('--share', required=False, action='store_true',
                    help='Tells if share the app', default=False)
parser.add_argument("--grid", type=int, help="Select the first grid to start from", default=None)
parser.add_argument("--run", type=int, help="Select the run in grid to start from", default=None)


def cli():
    args = parser.parse_args()
    exp_settings = dict(
        start_from_grid=args.grid,
        start_from_run=args.run,
        resume=args.resume,
        tracking_dir=args.dir
    )
    streamlit_entry(args.file, exp_settings, args.share)


if __name__ == '__main__':
    cli()



from ruamel.yaml import YAML


class ManipulationInputs:
    def __init__(self):
        self.updated_meta = None
        self.path = None
        self.filters = {}
        self.updated_config = None
        self.keys_to_delete = None
        self.files_to_delete = None
        self.artifacts_to_delete = None
        self.fix_string_params = None

    def update(self, path, filters, updated_config, updated_meta, to_delete, fix_string_params):
        if path:
            self.path = path

        if filters:
            self.filters = YAML(typ="safe", pure=True).load(filters)
        if updated_config:
            self.updated_config = YAML(typ="safe", pure=True).load(updated_config)
        if updated_meta:
            self.updated_meta = YAML(typ="safe", pure=True).load(updated_meta)

        if to_delete:
            dict_to_delete = YAML(typ="safe", pure=True).load(to_delete)
            self.keys_to_delete = dict_to_delete['keys']
            self.files_to_delete = dict_to_delete['files']
            self.artifacts_to_delete = dict_to_delete['artifacts']

        self.fix_string_params = fix_string_params



import torch
from torch import nn, Tensor
from torch.nn import functional as F
from typing import Optional


def _make_divisible(v: float, divisor: int, min_value: Optional[int] = None) -> int:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class ConvModule(nn.Sequential):
    def __init__(self, c1, c2, k, s=1, p=0, d=1, g=1):
        super().__init__(
            nn.Conv2d(c1, c2, k, s, p, d, g, bias=False),
            nn.BatchNorm2d(c2),
            nn.ReLU6(True)
        )


class SqueezeExcitation(nn.Module):
    def __init__(self, ch, squeeze_factor=4):
        super().__init__()
        squeeze_ch = _make_divisible(ch // squeeze_factor, 8)
        self.fc1 = nn.Conv2d(ch, squeeze_ch, 1)
        self.relu = nn.ReLU(True)
        self.fc2 = nn.Conv2d(squeeze_ch, ch, 1)

    def _scale(self, x: Tensor) -> Tensor:
        scale = F.adaptive_avg_pool2d(x, 1)
        scale = self.fc2(self.relu(self.fc1(scale)))
        return F.hardsigmoid(scale, True)

    def forward(self, x: Tensor) -> Tensor:
        scale = self._scale(x)
        return scale * x


class InvertedResidualConfig:
    def __init__(self, c1, c2, k, expanded_ch, use_se) -> None:
        pass


class InvertedResidual(nn.Module):
    def __init__(self, c1, c2, s, expand_ratio):
        super().__init__()
        ch = int(round(c1 * expand_ratio))
        self.use_res_connect = s == 1 and c1 == c2

        layers = []

        if expand_ratio != 1:
            layers.append(ConvModule(c1, ch, 1))

        layers.extend([
            ConvModule(ch, ch, 3, s, 1, g=ch),
            nn.Conv2d(ch, c2, 1, bias=False),
            nn.BatchNorm2d(c2)
        ])

        self.conv = nn.Sequential(*layers)
        
    def forward(self, x: Tensor) -> Tensor:
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


mobilenetv3_settings = {
    'S': [],
    'L': []
}


class MobileNetV3(nn.Module):
    def __init__(self, variant: str = None):
        super().__init__()
        self.out_indices = [3, 6, 13, 17]
        self.channels = [24, 32, 96, 320]
        input_channel = 32
        
        inverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        self.features = nn.ModuleList([ConvModule(3, input_channel, 3, 2, 1)])

        for t, c, n, s in inverted_residual_setting:
            output_channel = c
            for i in range(n):
                stride = s if i == 0 else 1
                self.features.append(InvertedResidual(input_channel, output_channel, stride, t))
                input_channel = output_channel
    
    def forward(self, x: Tensor) -> Tensor:
        outs = []
        for i, m in enumerate(self.features):
            x = m(x)
            if i in self.out_indices:
                outs.append(x)
        return outs


if __name__ == '__main__':
    model = MobileNetV3()
    # model.load_state_dict(torch.load('checkpoints/backbones/mobilenet_v2.pth', map_location='cpu'), strict=False)
    model.eval()
    x = torch.randn(1, 3, 224, 224)
    # outs = model(x)
    # for y in outs:
    #     print(y.shape)

    from fvcore.nn import flop_count_table, FlopCountAnalysis
    print(flop_count_table(FlopCountAnalysis(model, x)))


import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import ConvModule
from ezdl.models.modules import PPM


class AlignedModule(nn.Module):
    def __init__(self, c1, c2, k=3):
        super().__init__()
        self.down_h = nn.Conv2d(c1, c2, 1, bias=False)
        self.down_l = nn.Conv2d(c1, c2, 1, bias=False)
        self.flow_make = nn.Conv2d(c2 * 2, 2, k, 1, 1, bias=False)

    def forward(self, low_feature: Tensor, high_feature: Tensor) -> Tensor:
        high_feature_origin = high_feature
        H, W = low_feature.shape[-2:]
        low_feature = self.down_l(low_feature)
        high_feature = self.down_h(high_feature)
        high_feature = F.interpolate(high_feature, size=(H, W), mode='bilinear', align_corners=True)
        flow = self.flow_make(torch.cat([high_feature, low_feature], dim=1))
        high_feature = self.flow_warp(high_feature_origin, flow, (H, W))
        return high_feature

    def flow_warp(self, x: Tensor, flow: Tensor, size: tuple) -> Tensor:
        # norm = torch.tensor(size).reshape(1, 1, 1, -1)
        norm = torch.tensor([[[[*size]]]]).type_as(x).to(x.device)
        H = torch.linspace(-1.0, 1.0, size[0]).view(-1, 1).repeat(1, size[1])
        W = torch.linspace(-1.0, 1.0, size[1]).repeat(size[0], 1)
        grid = torch.cat((W.unsqueeze(2), H.unsqueeze(2)), dim=2)
        grid = grid.repeat(x.shape[0], 1, 1, 1).type_as(x).to(x.device)
        grid = grid + flow.permute(0, 2, 3, 1) / norm
        output = F.grid_sample(x, grid, align_corners=False)
        return output


class SFHead(nn.Module):
    def __init__(self, in_channels, channel=256, num_classes=19, scales=(1, 2, 3, 6)):
        super().__init__()
        self.ppm = PPM(in_channels[-1], channel, scales)

        self.fpn_in = nn.ModuleList([])
        self.fpn_out = nn.ModuleList([])
        self.fpn_out_align = nn.ModuleList([])

        for in_ch in in_channels[:-1]:
            self.fpn_in.append(ConvModule(in_ch, channel, 1))
            self.fpn_out.append(ConvModule(channel, channel, 3, 1, 1))
            self.fpn_out_align.append(AlignedModule(channel, channel//2))

        self.bottleneck = ConvModule(len(in_channels) * channel, channel, 3, 1, 1)
        self.dropout = nn.Dropout2d(0.1)
        self.conv_seg = nn.Conv2d(channel, num_classes, 1)

    def forward(self, features: list) -> Tensor:
        f = self.ppm(features[-1])
        fpn_features = [f]

        for i in reversed(range(len(features) - 1)):
            feature = self.fpn_in[i](features[i])
            f = feature + self.fpn_out_align[i](feature, f)
            fpn_features.append(self.fpn_out[i](f))

        fpn_features.reverse()

        for i in range(1, len(fpn_features)):
            fpn_features[i] = F.interpolate(fpn_features[i], size=fpn_features[0].shape[-2:], mode='bilinear', align_corners=True)

        output = self.bottleneck(torch.cat(fpn_features, dim=1))
        output = self.conv_seg(self.dropout(output))
        return output




import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import DropPath


class MLP(nn.Module):
    def __init__(self, dim, hidden_dim, out_dim=None) -> None:
        super().__init__()
        out_dim = out_dim or dim
        self.fc1 = nn.Linear(dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, out_dim)

    def forward(self, x: Tensor) -> Tensor:
        return self.fc2(F.gelu(self.fc1(x)))


class Attention(nn.Module):
    def __init__(self, dim, head, sr_ratio=1):
        super().__init__()
        self.head = head
        self.sr_ratio = sr_ratio 
        self.scale = (dim // head) ** -0.5

        self.q = nn.Linear(dim, dim)
        self.kv = nn.Linear(dim, dim*2)
        self.proj = nn.Linear(dim, dim)

        if sr_ratio > 1:
            self.sr = nn.Conv2d(dim, dim, sr_ratio+1, sr_ratio, sr_ratio//2, groups=dim)
            self.sr_norm = nn.LayerNorm(dim)

        self.apply_transform = head > 1
        if self.apply_transform:
            self.transform_conv = nn.Conv2d(head, head, 1, 1)
            self.transform_norm = nn.InstanceNorm2d(head)

    def forward(self, x: Tensor, H, W) -> Tensor:
        B, N, C = x.shape
        q = self.q(x).reshape(B, N, self.head, C // self.head).permute(0, 2, 1, 3)

        if self.sr_ratio > 1:
            x = x.permute(0, 2, 1).reshape(B, C, H, W)
            x = self.sr(x).reshape(B, C, -1).permute(0, 2, 1)
            x = self.sr_norm(x)
        k, v = self.kv(x).reshape(B, -1, 2, self.head, C // self.head).permute(2, 0, 3, 1, 4)

        attn = (q @ k.transpose(-2, -1)) * self.scale

        if self.apply_transform:
            attn = self.transform_conv(attn)
            attn = attn.softmax(dim=-1)
            attn = self.transform_norm(attn)
        else:
            attn = attn.softmax(dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        return x


class Block(nn.Module):
    def __init__(self, dim, head, sr_ratio=1, dpr=0.):
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.attn = Attention(dim, head, sr_ratio)
        self.drop_path = DropPath(dpr) if dpr > 0. else nn.Identity()
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, int(dim*4))

    def forward(self, x: Tensor, H, W) -> Tensor:
        x = x + self.drop_path(self.attn(self.norm1(x), H, W))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class PA(nn.Module):        # positional embedding
    def __init__(self, dim):
        super().__init__()
        self.pa_conv = nn.Conv2d(dim, dim, 3, 1, 1, groups=dim)
        
    def forward(self, x):
        return x * self.pa_conv(x).sigmoid()


class PatchEmbed(nn.Module):
    def __init__(self, c1=3, c2=64):
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, 3, 2, 1)
        self.norm = nn.BatchNorm2d(c2)
        self.pos = PA(c2)

    def forward(self, x: Tensor):
        x = self.pos(self.norm(self.conv(x)))
        _, _, H, W = x.shape
        x = x.flatten(2).transpose(1, 2)
        return x, H, W


class Stem(nn.Module):
    def __init__(self, c1=3, c2=64):
        super().__init__()
        ch = c2 // 2
        self.conv1 = nn.Conv2d(c1, ch, 3, 2, 1, bias=False)
        self.norm1 = nn.BatchNorm2d(ch)
        self.conv2 = nn.Conv2d(ch, ch, 3, 1, 1, bias=False)
        self.norm2 = nn.BatchNorm2d(ch)
        self.conv3 = nn.Conv2d(ch, c2, 3, 2, 1, bias=False)
        self.act = nn.ReLU()
        self.pos = PA(c2)

    def forward(self, x):
        x = self.act(self.norm1(self.conv1(x)))
        x = self.act(self.norm2(self.conv2(x)))
        x = self.pos(self.conv3(x))
        _, _, H, W = x.shape
        x = x.flatten(2).permute(0, 2, 1)
        return x, H, W


rest_settings = {
    'S': [[64, 128, 256, 512], [2, 2, 6, 2], 0.1],    # [embed_dims, depths, dpr]
    'B': [[96, 192, 384, 768], [2, 2, 6, 2], 0.2],
    'L': [[96, 192, 384, 768], [2, 2, 18, 2], 0.3]
}


class ResT(nn.Module):  
    def __init__(self, model_name: str = 'S') -> None:
        super().__init__()
        assert model_name in rest_settings.keys(), f"ResT model name should be in {list(rest_settings.keys())}"
        embed_dims, depths, drop_path_rate = rest_settings[model_name]
        self.channels = embed_dims
        self.stem = Stem(3, embed_dims[0])

        # patch_embed
        self.patch_embed_2 = PatchEmbed(embed_dims[0], embed_dims[1])
        self.patch_embed_3 = PatchEmbed(embed_dims[1], embed_dims[2])
        self.patch_embed_4 = PatchEmbed(embed_dims[2], embed_dims[3])

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]

        # transformer encoder
        cur = 0
        self.stage1 = nn.ModuleList([Block(embed_dims[0], 1, 8, dpr[cur+i]) for i in range(depths[0])])

        cur += depths[0]
        self.stage2 = nn.ModuleList([Block(embed_dims[1], 2, 4, dpr[cur+i]) for i in range(depths[1])])

        cur += depths[1]
        self.stage3 = nn.ModuleList([Block(embed_dims[2], 4, 2, dpr[cur+i]) for i in range(depths[2])])
        
        cur += depths[2]
        self.stage4 = nn.ModuleList([Block(embed_dims[3], 8, 1, dpr[cur+i]) for i in range(depths[3])])
        self.norm = nn.LayerNorm(embed_dims[-1])


    def forward(self, x: Tensor) -> Tensor:
        B = x.shape[0]
        
        # stage 1
        x, H, W = self.stem(x)
        for blk in self.stage1:
            x = blk(x, H, W)
        x1 = x.permute(0, 2, 1).reshape(B, -1, H, W)

        # stage 2
        x, H, W = self.patch_embed_2(x1)
        for blk in self.stage2:
            x = blk(x, H, W)
        x2 = x.permute(0, 2, 1).reshape(B, -1, H, W)

        # stage 3
        x, H, W = self.patch_embed_3(x2)
        for blk in self.stage3:
            x = blk(x, H, W)
        x3 = x.permute(0, 2, 1).reshape(B, -1, H, W)

        # stage 4
        x, H, W = self.patch_embed_4(x3)
        for blk in self.stage4:
            x = blk(x, H, W)
        x = self.norm(x)
        x4 = x.permute(0, 2, 1).reshape(B, -1, H, W)

        return x1, x2, x3, x4


if __name__ == '__main__':
    model = ResT('S')
    model.load_state_dict(torch.load('checkpoints/backbones/rest/rest_small.pth', map_location='cpu'), strict=False)
    x = torch.zeros(1, 3, 224, 224)
    outs = model(x)
    for y in outs:
        print(y.shape)


import torch
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import ConvModule


class CondHead(nn.Module):
    def __init__(self, in_channel: int = 2048, channel: int = 512, num_classes: int = 19):
        super().__init__()
        self.num_classes = num_classes
        self.weight_num = channel * num_classes
        self.bias_num = num_classes

        self.conv = ConvModule(in_channel, channel, 1)
        self.dropout = nn.Dropout2d(0.1)

        self.guidance_project = nn.Conv2d(channel, num_classes, 1)
        self.filter_project = nn.Conv2d(channel*num_classes, self.weight_num + self.bias_num, 1, groups=num_classes)

    def forward(self, features) -> Tensor:
        x = self.dropout(self.conv(features[-1]))
        B, C, H, W = x.shape
        guidance_mask = self.guidance_project(x)
        cond_logit = guidance_mask
        
        key = x
        value = x
        guidance_mask = guidance_mask.softmax(dim=1).view(*guidance_mask.shape[:2], -1)
        key = key.view(B, C, -1).permute(0, 2, 1)

        cond_filters = torch.matmul(guidance_mask, key)
        cond_filters /= H * W
        cond_filters = cond_filters.view(B, -1, 1, 1)
        cond_filters = self.filter_project(cond_filters)
        cond_filters = cond_filters.view(B, -1)

        weight, bias = torch.split(cond_filters, [self.weight_num, self.bias_num], dim=1)
        weight = weight.reshape(B * self.num_classes, -1, 1, 1)
        bias = bias.reshape(B * self.num_classes)

        value = value.view(-1, H, W).unsqueeze(0)
        seg_logit = F.conv2d(value, weight, bias, 1, 0, groups=B).view(B, self.num_classes, H, W)
        
        if self.training:
            return cond_logit, seg_logit
        return seg_logit


if __name__ == '__main__':
    from semseg.models.backbones import ResNetD
    backbone = ResNetD('50')
    head = CondHead()
    x = torch.randn(2, 3, 224, 224)
    features = backbone(x)
    outs = head(features)
    for out in outs:
        out = F.interpolate(out, size=x.shape[-2:], mode='bilinear', align_corners=False)
        print(out.shape)


import itertools
import os
from typing import Any, Union, Iterable, Mapping

import functools

import numpy as np
import torch
from PIL import Image
from torchvision.datasets.vision import VisionDataset
import torchvision.transforms as transforms
from torchvision.transforms import functional as F

from ezdl.transforms import \
    PairRandomCrop, SegOneHot, ToLong, FixValue, Denormalize, PairRandomFlip, squeeze0, \
    PairFlip, PairFourCrop
from .sequoia_stats import STATS as SEQUOIA_STATS
from .rededge_stats import STATS as REDEDGE_STATS
from sklearn.model_selection import train_test_split

from torch.utils.data.distributed import DistributedSampler
from super_gradients.training import utils as core_utils
from super_gradients.training.datasets.mixup import CollateMixup
from super_gradients.training.exceptions.dataset_exceptions import IllegalDatasetParameterException
from super_gradients.common.abstractions.abstract_logger import get_logger

from ezdl.data import DatasetInterface

logger = get_logger(__name__)


class WeedMapDatasetInterface(DatasetInterface):
    STATS = {
        'sequoia': SEQUOIA_STATS,
        'rededge': REDEDGE_STATS
    }

    def __init__(self, dataset_params, name=None):
        super(WeedMapDatasetInterface, self).__init__(dataset_params)
        channels = dataset_params['channels']
        if name == None:
            name = os.path.basename(os.path.normpath(dataset_params['root'])).lower()
        else:
            if name != os.path.basename(os.path.normpath(dataset_params.root)).lower():
                raise ValueError(f'Warning! Dataset name should match the dataset, found: {name}, {dataset_params.root}')
        self.dataset_name = name

        mean, std = self.get_mean_std(dataset_params['train_folders'], channels, name)

        self.lib_dataset_params = {
            'mean': mean,
            'std': std,
            'channels': channels
        }

        input_transform = [
            transforms.Normalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std']),
        ]

        test_transform = [
            transforms.Normalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std']),
        ]

        test_target_transform = [
            transforms.PILToTensor(),
            squeeze0,
            ToLong(),
            FixValue(source=10000, target=1),
            SegOneHot(num_classes=len(WeedMapDataset.CLASS_LABELS.keys()))
        ]

        target_transform = [
            transforms.PILToTensor(),
            squeeze0,
            ToLong(),
            FixValue(source=10000, target=1),
            SegOneHot(num_classes=len(WeedMapDataset.CLASS_LABELS.keys()))
        ]
        period = 1

        crop_size = core_utils.get_param(self.dataset_params, 'crop_size', default_val='same')
        if crop_size != 'same':
            crop = PairRandomCrop(crop_size)
            test_crop = PairFourCrop(crop_size, periodicity=period)

            input_transform.append(crop)
            target_transform.append(crop)
            test_transform.append(test_crop)
            test_target_transform.append(test_crop)
            period *= 4

        if dataset_params['hor_flip']:
            flip_hor = PairRandomFlip(orientation="horizontal")

            input_transform.append(flip_hor)
            target_transform.append(flip_hor)

        if dataset_params['ver_flip']:
            flip_ver = PairRandomFlip(orientation="vertical")

            input_transform.append(flip_ver)
            target_transform.append(flip_ver)

        if core_utils.get_param(self.dataset_params, 'size', default_val='same') != 'same':
            resize = transforms.Resize(size=core_utils.get_param(self.dataset_params, 'size', default_val='same'))

            input_transform.append(resize)
            target_transform.append(resize)
            test_transform.append(resize)
            test_target_transform.append(resize)

        target_transform = transforms.Compose(target_transform)
        input_transform = transforms.Compose(input_transform)
        test_transform = transforms.Compose(test_transform)
        test_target_transform = transforms.Compose(test_target_transform)

        # Divide train, val and test
        self.train_folders = dataset_params['train_folders']
        self.test_folders = dataset_params['test_folders']

        train_index = WeedMapDataset.build_index(self.dataset_params.root, self.train_folders, channels)
        test_index = WeedMapDataset.build_index(self.dataset_params.root, self.test_folders, channels)
        train_index, val_index = train_test_split(train_index, test_size=0.2)

        self.trainset = WeedMapDataset(root=self.dataset_params.root, channels=channels,
                                       batch_size=self.dataset_params.batch_size, index=train_index,
                                       transform=input_transform, target_transform=target_transform,
                                       return_path=dataset_params['return_path'])

        self.valset = WeedMapDataset(root=self.dataset_params.root, channels=channels,
                                     batch_size=self.dataset_params.val_batch_size, index=val_index,
                                     transform=input_transform, target_transform=target_transform,
                                     return_path=dataset_params['return_path'])

        self.testset = WeedMapDataset(root=self.dataset_params.root, channels=channels,
                                      batch_size=self.dataset_params.test_batch_size, index=test_index,
                                      transform=test_transform, target_transform=test_target_transform,
                                      return_path=dataset_params['return_path'], period=period)

    def undo_preprocess(self, x):
        return (Denormalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std'])(x) * 255).type(torch.uint8)

    @classmethod
    def get_mean_std(cls, train_folders, channels, dataset_name):
        stats = cls.STATS[dataset_name]
        if channels == 'CIR':
            channels = ['NIR', 'G', 'R']
        if len(train_folders) == 1:
            f = train_folders[0]
            return list(zip(*[(stats[f][c]['mean'], stats[f][c]['std']) for c in channels]))
        else:
            sums = {
                **{c + '_sum': sum([stats[f][c]['sum'] for f in train_folders]) for c in channels},
                **{c + '_sum_sq': sum([stats[f][c]['sum_sq'] for f in train_folders]) for c in channels}
            }
            count = sum([stats[f]['count'] for f in train_folders])
            means = [sums[c + '_sum'] / count for c in channels]
            stds = [np.sqrt((sums[c + '_sum_sq'] / count) - (means[i] ** 2))
                    for i, c in enumerate(channels)]
            return means, stds

    def get_run_loader(self, root=None, folders=None, batch_size=None):
        batch_size = batch_size if batch_size is not None else self.dataset_params.test_batch_size
        folders = folders if folders is not None else self.test_folders
        root = root if root is not None else self.dataset_params.root

        index = WeedMapDataset.build_index(root, folders, self.dataset_params.channels)
        input_transform = [
            transforms.Normalize(self.lib_dataset_params['mean'], self.lib_dataset_params['std']),
        ]

        target_transform = [
            transforms.PILToTensor(),
            squeeze0,
            ToLong(),
            FixValue(source=10000, target=1),
            SegOneHot(num_classes=len(WeedMapDataset.CLASS_LABELS.keys()))
        ]

        if core_utils.get_param(self.dataset_params, 'size', default_val='same') != 'same':
            resize = transforms.Resize(size=core_utils.get_param(self.dataset_params, 'size', default_val='same'))
            input_transform.append(resize)
            target_transform.append(resize)

        target_transform = transforms.Compose(target_transform)
        input_transform = transforms.Compose(input_transform)

        runset = WeedMapDataset(root=self.dataset_params.root, channels=self.dataset_params.channels,
                                batch_size=self.dataset_params.test_batch_size, index=index,
                                transform=input_transform, target_transform=target_transform)

        loader = torch.utils.data.DataLoader(runset,
                                             batch_size=batch_size,
                                             num_workers=self.dataset_params.num_workers,
                                             pin_memory=True,
                                             persistent_workers=False)
        return loader


class WeedMapDataset(VisionDataset):
    CLASS_LABELS = {0: "background", 1: "crop", 2: 'weed'}
    classes = ['background', 'crop', 'weed']

    def __init__(self,
                 root: str,
                 transform: callable,
                 target_transform: callable,
                 channels: Union[str, Iterable] = 'CIR',
                 index: Iterable = None,
                 batch_size: int = 4,
                 return_mask: bool = False,
                 return_path: bool = False,
                 period: int = None,
                 ):
        """
        Initialize a SequoiaDataset object.

        :param batch_size: The batch size.
        :param root: The root directory.
        :param transform: The transform to apply to the data.
        :param target_transform: The transform to apply to the target.
        :param channels: The channels to use.
        :param index: The index to get the images.
        :param return_mask: Whether to return the mask.
        :param return_path: Whether to return the path.
        :param period: Used when augmentation returns different augmentations for each image.
        """
        self.batch_size = batch_size
        super().__init__(root=root)

        if channels == 'CIR' or len(channels) == 1:
            channels = channels[0] if len(channels) == 1 else channels
            self.get_img = self.get_complete_image
        else:
            self.get_img = self.get_channels

        if index:
            self.index = index
        else:
            self.index = self.build_index(root, channels=channels)

        if period is not None:
            self.index = list(itertools.chain.from_iterable(itertools.repeat(x, period) for x in self.index))

        self.len = len(self.index)

        self.channels = channels
        self.n_files = {}
        self.path = root
        self.transform = transform
        self.target_transform = target_transform
        self.return_mask = return_mask
        self.return_name = return_path

    @classmethod
    def build_index(cls,
                    root: str,
                    macro_folders: Iterable = None,
                    channels: Union[Iterable, str] = 'CIR') -> list:

        if macro_folders is None:
            macro_folders = os.listdir(root)

        if channels == 'CIR':
            counter_channel = channels
        else:
            counter_channel = channels[0]  # Channel used to count images

        n_files = dict()
        for folder in macro_folders:
            n_files[folder] = (os.listdir(os.path.join(root, folder, 'tile', counter_channel)))
        index = dict()
        i = 0
        for folder, files in n_files.items():
            for file in files:
                index[i] = folder, file
                i += 1
        return list(index.values())

    def get_complete_image(self, folder, file) -> Any:
        img = Image.open(
            os.path.join(self.path, folder, 'tile', self.channels, file)
        )
        return F.to_tensor(img)

    def get_channels(self, folder, file) -> Any:
        return torch.stack([
            F.to_tensor(Image.open(
                os.path.join(self.path, folder, 'tile', c, file)
            )).squeeze(0) for c in self.channels
        ]
        )

    def __getitem__(self, index: int) -> Any:
        """
        Get the i-th image and related target

        :param idx: The index of the image.
        :return: A pair consisting of the image and the target
        """
        folder, file = self.index[index]
        img = self.get_img(folder, file)
        gt = Image.open(
            os.path.join(self.path, folder, 'groundtruth',
                         folder + '_' + file.split('.')[0] + '_GroundTruth_iMap.png'
                         )
        )
        img = self.transform(img)
        gt = self.target_transform(gt)

        if self.return_mask:
            mask = Image.open(
                os.path.join(self.path, folder, 'mask', file)
            )
            gt = (gt, mask)
        if self.return_name:
            return img, gt, {'input_name': folder + '_' + file}
        return img, gt

    def __len__(self) -> int:
        """
        Get the number of batches.

        :return: The number of batches.
        """
        return self.len



import functools
import json
import click
from easydict import EasyDict

from ezdl.utils.utilities import update_collection, load_yaml

import os
os.environ['CRASH_HANDLER'] = "FALSE"


@click.group()
def main():
    pass


def exp_options(f):
    options = [
          click.option("--resume", is_flag=True, help='Resume the experiment'),
          click.option("--file", default="", help='Set the config file'),
          click.option("--dir", default="", help='Set the local tracking directory'),
          click.option("--grid", type=int, default=0, help="Select the first grid to start from"),
          click.option("--run", type=int, default=0, help="Select the run in grid to start from"),
    ]
    return functools.reduce(lambda x, opt: opt(x), options, f)


@main.command("experiment")
@exp_options
def experiment(resume, file, dir, grid, run):
    from ezdl.experiment.experiment import experiment
    param_path = file or 'parameters.yaml'
    settings = load_yaml(param_path)
    settings['experiment'] = update_collection(settings['experiment'], resume, key='resume')
    settings['experiment'] = update_collection(settings['experiment'], grid, key='start_from_grid')
    settings['experiment'] = update_collection(settings['experiment'], run, key='start_from_run')
    settings['experiment'] = update_collection(settings['experiment'], dir, key='tracking_dir')
    experiment(settings)


@main.command("resume_run")
@click.option("--file", default="", help='Set the config resume file')
@click.option("--filters", type=json.loads, help="Filters to query in the resuming mode")
@click.option('-s', "--stage", type=json.loads, help="Stages to execute")
@click.option('-p', "--path", type=str, help="Path to the tracking url")
def resume_run(file, filters, stage, path):
    from ezdl.experiment.resume import resume_set_of_runs
    param_path = file or 'resume.yaml'
    settings = load_yaml(param_path)
    settings['runs'][0]['filters'] = update_collection(settings['runs'][0]['filters'], filters)
    settings['runs'][0]['stage'] = update_collection(settings['runs'][0]['stage'], stage)
    settings = update_collection(settings, path, key="path")
    resume_set_of_runs(settings)


@main.command("complete")
@click.option("--file", default="", help='Set the config resume file')
def complete(file):
    from ezdl.experiment.resume import complete_incompleted_runs
    param_path = file or 'resume.yaml'
    settings = load_yaml(param_path)
    complete_incompleted_runs(settings)


@main.command("manipulate")
def manipulate():
    from ezdl.wandb_manip import manipulate
    manipulate()


@main.command("app")
@exp_options
def app(resume, file, dir, grid, run):
    from ezdl.app import frontend
    args = EasyDict(
        resume=resume,
        file=file,
        dir=dir,
        grid=grid,
        run=run
    )
    frontend(args)




import os
import torch

import numpy as np

from typing import Union, Callable, Mapping, Any, List

from super_gradients.training.utils.callbacks import *
from super_gradients.training.utils.early_stopping import EarlyStop
from super_gradients.training.utils.utils import AverageMeter
from PIL import ImageColor, Image


def callback_factory(name, params, **kwargs):
    params = params or {}
    if name in ['early_stop', 'early_stopping', 'EarlyStop']:
        if "phase" in params:
            params.pop("phase")
        return EarlyStop(Phase.VALIDATION_EPOCH_END, **params)
    if name == "SegmentationVisualizationCallback":
        seg_trainer = kwargs['seg_trainer']
        loader = kwargs['loader']
        dataset = kwargs['dataset']
        params['freq'] = params.get('freq', 1)
        params['phase'] = Phase.VALIDATION_BATCH_END \
            if params['phase'] == 'validation' \
            else Phase.TEST_BATCH_END
        return SegmentationVisualizationCallback(logger=seg_trainer.sg_logger,
                                                 batch_idxs=[0, len(loader) - 1],
                                                 last_img_idx_in_batch=4,
                                                 num_classes=dataset.trainset.CLASS_LABELS,
                                                 undo_preprocessing=dataset.undo_preprocess,
                                                 **params)
    if params.get("phase"):
        params['phase'] = Phase.__dict__[params.get("phase")]
    return globals()[name](**params)


class SegmentationVisualizationCallback(PhaseCallback):
    test_sequence_name = 'test_seg'
    """
    A callback that adds a visualization of a batch of segmentation predictions to context.sg_logger
    Attributes:
        freq: frequency (in epochs) to perform this callback.
        batch_idx: batch index to perform visualization for.
        last_img_idx_in_batch: Last image index to add to log. (default=-1, will take entire batch).
    """

    def __init__(self, logger, phase: Phase, freq: int, num_classes, batch_idxs=None, last_img_idx_in_batch: int = None,
                 undo_preprocessing=None):
        super(SegmentationVisualizationCallback, self).__init__(phase)
        if batch_idxs is None:
            batch_idxs = [0]
        self.freq = freq
        self.num_classes = num_classes
        self.batch_idxs = batch_idxs
        self.last_img_idx_in_batch = last_img_idx_in_batch
        self.undo_preprocesing = undo_preprocessing
        self.prefix = 'train' if phase == Phase.TRAIN_EPOCH_END else 'val' \
            if phase == Phase.VALIDATION_BATCH_END else 'test'
        if phase == Phase.TEST_BATCH_END:
            logger.create_image_mask_sequence(f'{self.prefix}_seg')

    def __call__(self, context: PhaseContext):
        epoch = context.epoch if context.epoch is not None else 0
        if epoch % self.freq == 0 and context.batch_idx in self.batch_idxs:
            preds = context.preds.clone()
            SegmentationVisualization.visualize_batch(logger=context.sg_logger,
                                                      image_tensor=context.inputs,
                                                      pred_mask=preds,
                                                      target_mask=context.target,
                                                      num_classes=self.num_classes,
                                                      batch_name=context.batch_idx,
                                                      undo_preprocessing_func=self.undo_preprocesing,
                                                      prefix=self.prefix,
                                                      names=context.input_name,
                                                      iteration=context.epoch)
            if self.prefix == 'test' and context.batch_idx == self.batch_idxs[-1]:
                context.sg_logger.add_image_mask_sequence(f'{self.prefix}_seg')


class SegmentationVisualization:

    @staticmethod
    def _visualize_image(image_np: np.ndarray, pred_mask: torch.Tensor, target_mask: torch.Tensor, classes):
        """

        :param image_np: numpy image
        :param pred_mask: (C, H, W) tensor of classes in one hot encoding
        :param target_mask: (H, W) tensor of classes
        :param classes:
        :return:
        """
        pred_mask = torch.tensor(pred_mask.copy())
        target_mask = torch.tensor(target_mask.copy())

        pred_mask = pred_mask.argmax(dim=0)

        if image_np.shape[0] < 3:
            image_np = torch.vstack([image_np,
                                     torch.zeros((3 - image_np.shape[0], *image_np.shape[1:]), dtype=torch.uint8)]
                                    )
        image_np = image_np[:3, :, :]  # Take only 3 bands if there are more
        image_np = np.moveaxis(image_np.numpy(), 0, -1)

        return image_np, {
            "predictions": {
                "mask_data": pred_mask.numpy(),
                "class_labels": classes,
            },
            "ground_truth": {
                "mask_data": target_mask.numpy(),
                "class_labels": classes,
            },
        }

    @staticmethod
    def visualize_batch(logger, image_tensor: torch.Tensor, pred_mask: torch.Tensor, target_mask: torch.Tensor,
                        num_classes,
                        batch_name: Union[int, str],
                        undo_preprocessing_func: Callable[[torch.Tensor], np.ndarray] = lambda x: x,
                        image_scale: float = 1.,
                        prefix: str = '',
                        names: List[str] = None,
                        iteration: int = 0):
        """
        A helper function to visualize detections predicted by a network:
        saves images into a given path with a name that is {batch_name}_{imade_idx_in_the_batch}.jpg, one batch per call.
        Colors are generated on the fly: uniformly sampled from color wheel to support all given classes.

        :param iteration:
        :param names:
        :param prefix:
        :param image_tensor:            rgb images, (B, H, W, 3)
        :param batch_name:              id of the current batch to use for image naming
        :param undo_preprocessing_func: a function to convert preprocessed images tensor into a batch of cv2-like images
        :param image_scale:             scale factor for output image
        """
        image_np = undo_preprocessing_func(image_tensor.detach()).type(dtype=torch.uint8).cpu()

        if names is None:
            names = ['_'.join([prefix, 'seg', str(batch_name), str(i)]) if prefix == 'val' else \
                         '_'.join([prefix, 'seg', str(batch_name * image_np.shape[0] + i)]) for i in
                     range(image_np.shape[0])]
        else:
            names = [f"{prefix}_seg_{name}" for name in names]

        for i in range(image_np.shape[0]):
            preds = pred_mask[i].detach().cpu().numpy()
            targets = target_mask[i].detach().cpu().numpy()

            img, mask_dict = SegmentationVisualization._visualize_image(image_np[i], preds, targets, num_classes)
            if prefix == 'val':
                logger.add_mask(names[i], img, mask_dict, global_step=iteration)
            else:
                logger.add_image_mask_to_sequence(SegmentationVisualizationCallback.test_sequence_name,
                                                  names[i], img, mask_dict)


# class MlflowCallback(PhaseCallback):
#     """
#     A callback that logs metrics to MLFlow.
#     """
#
#     def __init__(self, phase: Phase, freq: int,
#                  client: MLRun,
#                  params: Mapping = None
#                  ):
#         """
#         param phase: phase to log metrics for
#         param freq: frequency of logging
#         param client: MLFlow client
#         """
#
#         if phase == Phase.TRAIN_EPOCH_END:
#             self.prefix = 'train_'
#         elif phase == Phase.VALIDATION_EPOCH_END:
#             self.prefix = 'val_'
#         else:
#             raise NotImplementedError('Unrecognized Phase')
#
#         super(MlflowCallback, self).__init__(phase)
#         self.freq = freq
#         self.client = client
#
#         if params:
#             self.client.log_params(params)
#
#     def __call__(self, context: PhaseContext):
#         """
#         Logs metrics to MLFlow.
#             param context: context of the current phase
#         """
#         if context.epoch % self.freq == 0:
#             self.client.log_metrics({self.prefix + k: v for k, v in context.metrics_dict.items()})


class MetricsLogCallback(PhaseCallback):
    """
    A callback that logs metrics to MLFlow.
    """

    def __init__(self, phase: Phase, freq: int
                 ):
        """
        param phase: phase to log metrics for
        param freq: frequency of logging
        param client: MLFlow client
        """

        if phase == Phase.TRAIN_EPOCH_END:
            self.prefix = 'train_'
        elif phase == Phase.VALIDATION_EPOCH_END:
            self.prefix = 'val_'
        else:
            raise NotImplementedError('Unrecognized Phase')

        super(MetricsLogCallback, self).__init__(phase)
        self.freq = freq

    def __call__(self, context: PhaseContext):
        """
        Logs metrics to MLFlow.
            param context: context of the current phase
        """
        if self.phase == Phase.TRAIN_EPOCH_END:
            context.sg_logger.add_scalar('epoch', context.epoch)
        if context.epoch % self.freq == 0:
            context.sg_logger.add_scalars({self.prefix + k: v for k, v in context.metrics_dict.items()})


class AverageMeterCallback(PhaseCallback):
    def __init__(self):
        super(AverageMeterCallback, self).__init__(Phase.TEST_BATCH_END)
        self.meters = {}

    def __call__(self, context: PhaseContext):
        """
        Logs metrics to MLFlow.
            param context: context of the current phase
        """
        context.metrics_compute_fn.update(context.preds, context.target)
        metrics_dict = context.metrics_compute_fn.compute()
        for k, v in metrics_dict.items():
            if not self.meters.get(k):
                self.meters[k] = AverageMeter()
            self.meters[k].update(v, 1)


class SaveSegmentationPredictionsCallback(PhaseCallback):
    def __init__(self, phase, path, num_classes):
        super(SaveSegmentationPredictionsCallback, self).__init__(phase)
        self.path = path
        self.num_classes = num_classes

        os.makedirs(self.path, exist_ok=True)
        colors = ['blue', 'green', 'red']
        self.colors = []
        for color in colors:
            if isinstance(color, str):
                color = ImageColor.getrgb(color)
            self.colors.append(torch.tensor(color, dtype=torch.uint8))

    def __call__(self, context: PhaseContext):
        for prediction, input_name in zip(context.preds, context.input_name):
            path = os.path.join(self.path, input_name)
            prediction = prediction.detach().cpu()
            masks = torch.concat([
                (prediction.argmax(0) == cls).unsqueeze(0)
                for cls in range(self.num_classes)
            ])

            img_to_draw = torch.zeros(*prediction.shape[-2:], 3, dtype=torch.uint8)
            # TODO: There might be a way to vectorize this
            for mask, color in zip(masks, self.colors):
                img_to_draw[mask] = color

            Image.fromarray(img_to_draw.numpy()).save(path)



import json
import os
import signal
import time
from typing import Union, Any

import psutil
import torch
from PIL import Image
from flatbuffers.builder import np
from matplotlib import pyplot as plt
from super_gradients.common import ADNNModelRepositoryDataInterfaces
from super_gradients.common.environment.ddp_utils import multi_process_safe
from super_gradients.common.sg_loggers.abstract_sg_logger import AbstractSGLogger
from super_gradients.training.params import TrainingParams
from super_gradients.training.utils import sg_trainer_utils
from super_gradients.common.abstractions.abstract_logger import get_logger


logger = get_logger(__name__)


class BaseSGLogger(AbstractSGLogger):

    def __init__(self, project_name: str,
                 experiment_name: str,
                 storage_location: str,
                 resumed: bool,
                 training_params: TrainingParams,
                 checkpoints_dir_path: str,
                 tb_files_user_prompt: bool = False,
                 launch_tensorboard: bool = False,
                 tensorboard_port: int = None,
                 save_checkpoints_remote: bool = True,
                 save_tensorboard_remote: bool = True,
                 save_logs_remote: bool = True):
        """

        :param experiment_name: Used for logging and loading purposes
        :param storage_location: If set to 's3' (i.e. s3://my-bucket) saves the Checkpoints in AWS S3 otherwise saves the Checkpoints Locally
        :param resumed: if true, then old tensorboard files will *not* be deleted when tb_files_user_prompt=True
        :param training_params: training_params for the experiment.
        :param checkpoints_dir_path: Local root directory path where all experiment logging directories will
                                                 reside.
        :param tb_files_user_prompt: Asks user for Tensorboard deletion prompt.
        :param launch_tensorboard: Whether to launch a TensorBoard process.
        :param tensorboard_port: Specific port number for the tensorboard to use when launched (when set to None, some free port
                    number will be used
        :param save_checkpoints_remote: Saves checkpoints in s3.
        :param save_tensorboard_remote: Saves tensorboard in s3.
        :param save_logs_remote: Saves log files in s3.
        """
        super().__init__()
        self.sequences = {}
        self.project_name = project_name
        self.experiment_name = experiment_name
        self.storage_location = storage_location

        if storage_location.startswith('s3'):
            self.save_checkpoints_remote = save_checkpoints_remote
            self.save_tensorboard_remote = save_tensorboard_remote
            self.save_logs_remote = save_logs_remote
            self.remote_storage_available = True
        else:
            self.remote_storage_available = False
            if save_checkpoints_remote:
                logger.error('save_checkpoints_remote == True but storage_location is not s3 path. Files will not be saved remotely')
            if save_tensorboard_remote:
                logger.error('save_tensorboard_remote == True but storage_location is not s3 path. Files will not be saved remotely')
            if save_logs_remote:
                logger.error('save_logs_remote == True but storage_location is not s3 path. Files will not be saved remotely')

            self.save_checkpoints_remote = False
            self.save_tensorboard_remote = False
            self.save_logs_remote = False

        self.tensor_board_process = None
        self.max_global_steps = training_params.max_epochs
        self._local_dir = checkpoints_dir_path

        self._make_dir()
        self._init_log_file()

        self.model_checkpoints_data_interface = ADNNModelRepositoryDataInterfaces(data_connection_location=self.storage_location)

        self.use_tensorboard = launch_tensorboard
        if launch_tensorboard:
            self._launch_tensorboard(port=tensorboard_port)
            self._init_tensorboard(resumed, tb_files_user_prompt)

    @multi_process_safe
    def _launch_tensorboard(self, port):
        self.tensor_board_process, _ = sg_trainer_utils.launch_tensorboard_process(self._local_dir, port=port)

    @multi_process_safe
    def _init_tensorboard(self, resumed, tb_files_user_prompt):
        self.tensorboard_writer = sg_trainer_utils.init_summary_writer(self._local_dir, resumed, tb_files_user_prompt)

    @multi_process_safe
    def _make_dir(self):
        if not os.path.isdir(self._local_dir):
            os.makedirs(self._local_dir)

    @multi_process_safe
    def _init_log_file(self):
        time_string = time.strftime('%b%d_%H_%M_%S', time.localtime())
        self.log_file_path = f'{self._local_dir}/log_{time_string}.txt'

    @multi_process_safe
    def _write_to_log_file(self, lines: list):
        with open(self.log_file_path, 'a' if os.path.exists(self.log_file_path) else 'w') as log_file:
            for line in lines:
                log_file.write(line + '\n')

    @multi_process_safe
    def add_config(self, tag: str, config: dict):
        log_lines = ['--------- config parameters ----------']
        log_lines.append(json.dumps(config, indent=4, default=str))
        log_lines.append('------- config parameters end --------')

        if self.use_tensorboard:
            self.tensorboard_writer.add_text("Hyper_parameters", json.dumps(config, indent=4, default=str)
                                             .replace(" ", "&nbsp;").replace("\n", "  \n  "))
        self._write_to_log_file(log_lines)

    @multi_process_safe
    def add_scalar(self, tag: str, scalar_value: float, global_step: int = None):
        if self.use_tensorboard:
            self.tensorboard_writer.add_scalar(tag=tag.lower().replace(' ', '_'), scalar_value=scalar_value, global_step=global_step)

    @multi_process_safe
    def add_scalars(self, tag_scalar_dict: dict, global_step: int = None):
        """
        add multiple scalars.
        Unlike Tensorboard implementation, this does not add all scalars with a main tag (all scalars to the same chart).
        Instead, scalars are added to tensorboard like in add_scalar and are written in log together.
        """
        if self.use_tensorboard:
            for tag, value in tag_scalar_dict.items():
                self.tensorboard_writer.add_scalar(tag=tag.lower().replace(' ', '_'), scalar_value=value, global_step=global_step)

            self.tensorboard_writer.flush()

        # WRITE THE EPOCH RESULTS TO LOG FILE
        log_line = f'\nEpoch ({global_step}/{self.max_global_steps})  - '
        for tag, value in tag_scalar_dict.items():
            if isinstance(value, torch.Tensor):
                value = value.item()
            log_line += f'{tag.replace(" ", "_")}: {value}\t'

        self._write_to_log_file([log_line])

    @multi_process_safe
    def add_image(self, tag: str, image: Union[torch.Tensor, np.array, Image.Image], data_format='CHW', global_step: int = None):
        if self.use_tensorboard:
            self.tensorboard_writer.add_image(tag=tag, img_tensor=image, dataformats=data_format, global_step=global_step)

    @multi_process_safe
    def add_images(self, tag: str, images: Union[torch.Tensor, np.array], data_format='NCHW', global_step: int = None):
        """
        Add multiple images to SGLogger.
        Typically, this function will add a set of images to tensorboard, save them to disk or add it to experiment management framework.

        :param tag: Data identifier
        :param images: images to be added. The values should lie in [0, 255] for type uint8 or [0, 1] for type float.
        :param data_format: Image data format specification of the form NCHW, NHWC, CHW, HWC, HW, WH, etc.
        :param global_step: Global step value to record
        """
        if self.use_tensorboard:
            self.tensorboard_writer.add_images(tag=tag, img_tensor=images, dataformats=data_format, global_step=global_step)

    @multi_process_safe
    def add_video(self, tag: str, video: Union[torch.Tensor, np.array], global_step: int = None):
        """
        Add a single video to SGLogger.
        Typically, this function will add a video to tensorboard, save it to disk or add it to experiment management framework.

        :param tag: Data identifier
        :param video: the video to add. shape (N,T,C,H,W) or (T,C,H,W). The values should lie in [0, 255] for type uint8 or [0, 1] for type float.
        :param global_step: Global step value to record
        """
        if video.ndim < 5:
            video = video[None, ]
        if self.use_tensorboard:
            self.tensorboard_writer.add_video(tag=tag, video=video, global_step=global_step)

    @multi_process_safe
    def add_histogram(self, tag: str, values: Union[torch.Tensor, np.array], bins: str, global_step: int = None):
        if self.use_tensorboard:
            self.tensorboard_writer.add_histogram(tag=tag, values=values, bins=bins, global_step=global_step)

    @multi_process_safe
    def add_model_graph(self, tag: str, model: torch.nn.Module, dummy_input: torch.Tensor):
        """
        Add a pytorch model graph to the SGLogger.
        Only the model structure/architecture will be preserved and collected, NOT the model weights.

        :param tag: Data identifier
        :param model: the model to be added
        :param dummy_input: an input to be used for a forward call on the model
        """
        if self.use_tensorboard:
            self.tensorboard_writer.add_graph(model=model, input_to_model=dummy_input)

    @multi_process_safe
    def add_text(self, tag: str, text_string: str, global_step: int = None):
        if self.use_tensorboard:
            self.tensorboard_writer.add_text(tag=tag, text_string=text_string, global_step=global_step)

    @multi_process_safe
    def add_figure(self, tag: str, figure: plt.figure, global_step: int = None):
        """
        Add a text to SGLogger.
        Typically, this function will add a figure to tensorboard or add it to experiment management framework.

        :param tag: Data identifier
        :param figure: the figure to add
        :param global_step: Global step value to record
        """
        if self.use_tensorboard:
            self.tensorboard_writer.add_figure(tag=tag, figure=figure, global_step=global_step)

    @multi_process_safe
    def add_file(self, file_name: str = None):
        if self.remote_storage_available and self.use_tensorboard:
            self.model_checkpoints_data_interface.save_remote_tensorboard_event_files(self.experiment_name, self._local_dir, file_name)

    @multi_process_safe
    def upload(self):
        if self.save_tensorboard_remote and self.use_tensorboard:
            self.model_checkpoints_data_interface.save_remote_tensorboard_event_files(self.experiment_name, self._local_dir)

        if self.save_logs_remote:
            log_file_name = self.log_file_path.split('/')[-1]
            self.model_checkpoints_data_interface.save_remote_checkpoints_file(self.experiment_name, self._local_dir, log_file_name)

    @multi_process_safe
    def flush(self):
        if self.use_tensorboard:
            self.tensorboard_writer.flush()

    @multi_process_safe
    def close(self):
        if self.use_tensorboard:
            self.tensorboard_writer.close()
        if self.tensor_board_process is not None:
            try:
                logger.info('[CLEANUP] - Stopping tensorboard process')
                process = psutil.Process(self.tensor_board_process.pid)
                process.send_signal(signal.SIGTERM)
                logger.info('[CLEANUP] - Successfully stopped tensorboard process')
            except Exception as ex:
                logger.info('[CLEANUP] - Could not stop tensorboard process properly: ' + str(ex))

    @multi_process_safe
    def add_checkpoint(self, tag: str, state_dict: dict, global_step: int = None):

        name = f'ckpt_{global_step}.pth' if tag is None else tag
        if not name.endswith('.pth'):
            name += '.pth'

        path = os.path.join(self._local_dir, name)
        torch.save(state_dict, path)
        if tag == 'ckpt_best.pth':
            logger.info("Checkpoint saved in " + path)
        if self.save_checkpoints_remote:
            self.model_checkpoints_data_interface.save_remote_checkpoints_file(self.experiment_name, self._local_dir, name)

    @multi_process_safe
    def add_plot(self, tag: str, values, xtitle, ytitle, classes_marker=None):
        pass

    def add(self, tag: str, obj: Any, global_step: int = None):
        pass

    def local_dir(self) -> str:
        return self._local_dir


from itertools import groupby, product
from typing import Mapping

import collections


def linearize(dictionary: Mapping):
    """
    Linearize a nested dictionary making keys, tuples
    :param dictionary: nested dict
    :return: one level dict
    """
    exps = []
    for key, value in dictionary.items():
        if isinstance(value, collections.abc.Mapping):
            exps.extend(((key, lin_key), lin_value) for lin_key, lin_value in linearize(value))
        elif isinstance(value, list):
            exps.append((key, value))
        elif value is None:
            exps.append((key, [{}]))
        else:
            raise ValueError(f"Only dict, lists or None!!! -> {value} is {type(value)} for key {key}")
    return exps


def linearized_to_string(lin_dict):
    def linearize_key(key):
        if type(key) == tuple:
            return f"{key[0]}.{linearize_key(key[1])}"
        return key

    return [(linearize_key(key), value) for key, value in lin_dict]


def extract(elem: tuple):
    """
    Exctract the element of a single element tuple
    :param elem: tuple
    :return: element of the tuple if singleton or the tuple itself
    """
    if len(elem) == 1:
        return elem[0]
    return elem


def delinearize(lin_dict):
    """
    Convert a dictionary where tuples can be keys in na nested dictionary
    :param lin_dict: dicionary where keys can be tuples
    :return:
    """
    # Take keys that are tuples
    filtered = list(filter(lambda x: isinstance(x[0], tuple), lin_dict.items()))
    # Group it to make one level
    grouped = groupby(filtered, lambda x: x[0][0])
    # Create the new dict and apply recursively
    new_dict = {k: delinearize({extract(elem[0][1:]): elem[1] for elem in v}) for k, v in grouped}
    # Remove old items and put new ones
    for key, value in filtered:
        lin_dict.pop(key)
    delin_dict = {**lin_dict, **new_dict}
    return delin_dict


def make_grid(dict_of_list, return_cartesian_elements=False):
    """
    Produce a list of dict for each combination of values in the input dict given by the list of values
    :param return_cartesian_elements: if True return the elements that differs from the base dict
    :param dict_of_list: a dictionary where values can be lists
    :params return_cartesian_elements: return elements multiplied

    :return: a list of dictionaries given by the cartesian product of values in the input dictionary
    """
    # Linearize the dict to make the cartesian product straight forward
    linearized_dict = linearize(dict_of_list)
    # Compute the grid
    keys, values = zip(*linearized_dict)
    if any(map(lambda x: len(x) == 0, values)):
        raise ValueError("There shouldn't be empty lists in grid!!")
    grid_dict = list(dict(zip(keys, values_list)) for values_list in product(*values))
    # Delinearize the list of dicts
    grid = [delinearize(dictionary) for dictionary in grid_dict]
    if return_cartesian_elements:
        ce = list(filter(lambda x: len(x[1]) > 1, linearized_dict))
        return grid, ce
    return grid



import datetime
import os

import numpy as np
import wandb
import pandas as pd

from ezdl.experiment.run import Run

from super_gradients.common.abstractions.abstract_logger import get_logger

logger = get_logger(__name__)


def resume_set_of_runs(settings, post_filters=None):
    queries = settings['runs']
    path = settings['path']
    for query in queries:
        filters = query['filters']
        stage = query['stage']
        updated_config = query['updated_config']
        api = wandb.Api()
        runs = api.runs(path=path, filters=filters)
        runs = list(filter(post_filters, runs))
        print("Runs to resume:")
        for run in runs:
            print(f"{run.group} \t - \t {run.name}")
        if len(runs) == 0:
            logger.error(f'No runs found for query {filters} with post_filters: {post_filters}')
        for run in runs:
            resume_run(run, updated_config, stage)


def complete_incompleted_runs(settings):
    print("Going on to complete runs!")
    resume_set_of_runs(settings, lambda x: 'f1' not in x.summary)


def resume_run(wandb_run, updated_config, stage):
    to_resume_run = Run()
    to_resume_run.resume(wandb_run=wandb_run, updated_config=updated_config, phases=stage)
    to_resume_run.launch()


def get_interrupted_run(input_settings):
    namespace = input_settings["name"]
    group = input_settings["group"]
    last_run = wandb.Api().runs(path=namespace, filters={"group": group,}, order="-created_at")[0]
    filters = {"group": group, "name": last_run.id}
    stage = ["train", "test"]
    updated_config = None
    api = wandb.Api()
    runs = api.runs(path=namespace, filters=filters)
    if len(runs) == 0:
        raise RuntimeError("No runs found")
    if len(runs) > 1:
        raise EnvironmentError("More than 1 run???")
    to_resume_run = Run()
    to_resume_run.resume(wandb_run=runs[0], updated_config=updated_config, phases=stage)
    return to_resume_run


def retrieve_run_to_resume(settings, grids):
    grid_list = [(i, j) for i in range(len(grids)) for j in range(len(grids[i]))]
    dir = settings['tracking_dir']
    exp_log = ExpLog(track_dir=dir, wandb_path=settings['name'], group=settings['group'])
    i, j, finished = exp_log.get_last_run()
    if i is None:
        return 0, 0, False
    index = grid_list.index((i, j))
    try:
        start_grid, start_run = grid_list[index + 1]  # Skip interrupted run
    except IndexError as e:
        if finished:
            logger.info(e)
            raise ValueError('No experiment to resume!!')
        else:
            return len(grids), None, True
    resume_last = not finished
    return start_grid, start_run, resume_last


class ExpLog:
    STARTED = 'started'
    FINISHED = 'finished'
    CRASHED = 'crashed'
    FORMAT = "%d-%m-%YT%H:%M:%S"
    FILENAME = 'exp_log.csv'

    def __init__(self, track_dir, wandb_path, group):
        """

        :param track_dir: Directory where experiments are saves
        :param wandb_path: wand path
        :param group: wandb group
        """
        self.path = os.path.join(track_dir if track_dir is not None else '', self.FILENAME)
        self.wandb_path = wandb_path
        self.group = group
        try:
            self.exp_log = pd.read_csv(self.path)
        except FileNotFoundError:
            self.exp_log = pd.DataFrame(columns=['path', 'group', 'grid', 'run', 'crashed', 'started', 'finished'])

    def get_run_condition(self, grid, run):
        return (
                (self.exp_log.path == self.wandb_path) &
                (self.exp_log.group == self.group) &
                (self.exp_log.grid == grid) &
                (self.exp_log.run == run)
        )

    def cur_time(self):
        return datetime.datetime.now().strftime(self.FORMAT)

    def _insert_run(self, grid, run, started=True, finished=False):
        run_row = pd.DataFrame({
            'path': [self.wandb_path],
            'group': [self.group],
            'grid': [grid],
            'run': [run],
            'crashed': [False],
            'started': [self.cur_time() if started else None],
            'finished': [self.cur_time() if finished else None]
        })
        self.exp_log = pd.concat([self.exp_log, run_row])
        self.exp_log = self.exp_log.reset_index(drop=True)

    def insert_run(self, grid, run):
        run_condition = self.get_run_condition(grid, run)
        run_row = self.exp_log[run_condition]
        if run_row.empty:
            self._insert_run(grid, run)
        else:
            self.exp_log.loc[self.exp_log[run_condition].index, "started"] = self.cur_time()
            self.exp_log.loc[self.exp_log[run_condition].index, "crashed"] = False
            self.exp_log.loc[self.exp_log[run_condition].index, "finished"] = None

        self.save()

    def finish_run(self, grid, run, crashed=False):
        run_condition = self.get_run_condition(grid, run)
        run_row = self.exp_log[run_condition]
        if run_row.empty:
            logger.warning("Finishing a non started run on the CSV")
            self._insert_run(grid, run, finished=True)
        else:
            self.exp_log.loc[self.exp_log[run_condition].index, "crashed"] = crashed
            self.exp_log.loc[self.exp_log[run_condition].index, "finished"] = self.cur_time()
        self.save()

    def get_last_run(self):
        cur_exp_dataframe = self.exp_log[
            (self.exp_log.path == self.wandb_path) &
            (self.exp_log.group == self.group)
        ]
        last_run = cur_exp_dataframe[(cur_exp_dataframe.grid == cur_exp_dataframe.grid.max())].reset_index(drop=True)
        last_run = last_run[(last_run.run == last_run.run.max())]
        if len(last_run) > 1:
            raise Exception("More than one last run")
        elif len(last_run) == 0:
            logger.warning("This experiment is never started, no runs to resume")
            return None, None, False
        last_run = last_run.reset_index(drop=True).loc[0]  # Get the Series
        return last_run.grid, last_run.run, not pd.isnull(last_run.finished) and not last_run.crashed

    def save(self):
        self.exp_log.to_csv(self.path, index=False)

    def __repr__(self):
        return f"Local path : {self.path} \n" \
               f"Wandb path : {self.wandb_path} \n" \
               f"Wanfb group: {self.group} \n" + \
            self.exp_log.__repr__()



import torch
from torch import nn, Tensor
from ezdl.models.layers import DropPath


class PatchEmbed(nn.Module):
    """Image to Patch Embedding with overlapping
    """
    def __init__(self, patch_size=16, stride=16, padding=0, in_ch=3, embed_dim=768):
        super().__init__()
        self.proj = nn.Conv2d(in_ch, embed_dim, patch_size, stride, padding)

    def forward(self, x: torch.Tensor) -> Tensor:
        x = self.proj(x)                   # b x hidden_dim x 14 x 14
        return x


class Pooling(nn.Module):
    def __init__(self, pool_size=3) -> None:
        super().__init__()
        self.pool = nn.AvgPool2d(pool_size, 1, pool_size//2, count_include_pad=False)
    
    def forward(self, x: Tensor) -> Tensor:
        return self.pool(x) - x


class MLP(nn.Module):
    def __init__(self, dim, hidden_dim, out_dim=None) -> None:
        super().__init__()
        out_dim = out_dim or dim
        self.fc1 = nn.Conv2d(dim, hidden_dim, 1)
        self.act = nn.GELU()
        self.fc2 = nn.Conv2d(hidden_dim, out_dim, 1)

    def forward(self, x: Tensor) -> Tensor:
        return self.fc2(self.act(self.fc1(x)))


class PoolFormerBlock(nn.Module):
    def __init__(self, dim, pool_size=3, dpr=0., layer_scale_init_value=1e-5):
        super().__init__()
        self.norm1 = nn.GroupNorm(1, dim)
        self.token_mixer = Pooling(pool_size)
        self.norm2 = nn.GroupNorm(1, dim)
        self.drop_path = DropPath(dpr) if dpr > 0. else nn.Identity()
        self.mlp = MLP(dim, int(dim*4))
        
        self.layer_scale_1 = nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
        self.layer_scale_2 = nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)

    def forward(self, x: Tensor) -> Tensor:
        x = x + self.drop_path(self.layer_scale_1.unsqueeze(-1).unsqueeze(-1) * self.token_mixer(self.norm1(x))) 
        x = x + self.drop_path(self.layer_scale_2.unsqueeze(-1).unsqueeze(-1) * self.mlp(self.norm2(x))) 
        return x

poolformer_settings = {
    'S24': [[4, 4, 12, 4], [64, 128, 320, 512], 0.1],       # [layers, embed_dims, drop_path_rate]
    'S36': [[6, 6, 18, 6], [64, 128, 320, 512], 0.2],
    'M36': [[6, 6, 18, 6], [96, 192, 384, 768], 0.3]
}


class PoolFormer(nn.Module):     
    def __init__(self, model_name: str = 'S24') -> None:
        super().__init__()
        assert model_name in poolformer_settings.keys(), f"PoolFormer model name should be in {list(poolformer_settings.keys())}"
        layers, embed_dims, drop_path_rate = poolformer_settings[model_name]
        self.channels = embed_dims
    
        self.patch_embed = PatchEmbed(7, 4, 2, 3, embed_dims[0])

        network = []

        for i in range(len(layers)):
            blocks = []
            for j in range(layers[i]):
                dpr = drop_path_rate * (j + sum(layers[:i])) / (sum(layers) - 1)
                blocks.append(PoolFormerBlock(embed_dims[i], 3, dpr))

            network.append(nn.Sequential(*blocks))
            if i >= len(layers) - 1: break
            network.append(PatchEmbed(3, 2, 1, embed_dims[i], embed_dims[i+1]))

        self.network = nn.ModuleList(network)

        self.out_indices = [0, 2, 4, 6]
        for i, index in enumerate(self.out_indices):
            self.add_module(f"norm{index}", nn.GroupNorm(1, embed_dims[i]))

    def forward(self, x: Tensor):
        x = self.patch_embed(x)
        outs = []

        for i, blk in enumerate(self.network):
            x = blk(x)

            if i in self.out_indices:
                out = getattr(self, f"norm{i}")(x)
                outs.append(out)
        return outs
        

if __name__ == '__main__':
    model = PoolFormer('S24')
    model.load_state_dict(torch.load('C:\\Users\\sithu\\Documents\\weights\\backbones\\poolformer\\poolformer_s24.pth.tar', map_location='cpu'), strict=False)
    x = torch.randn(1, 3, 224, 224)
    feats = model(x)
    for y in feats:
        print(y.shape)



import torch
from torch import nn, Tensor
from torch.nn import functional as F
from einops import rearrange

RATIOS = [8, 4, 2]


class MLP(nn.Module):
    def __init__(self, dim=2048, embed_dim=768):
        super().__init__()
        self.proj = nn.Linear(dim, embed_dim)

    def forward(self, x: Tensor) -> Tensor:
        x = x.flatten(2).transpose(1, 2)
        x = self.proj(x)
        return x


class PatchEmbed(nn.Module):
    def __init__(self, patch_size=4, in_ch=3, dim=96, type='pool') -> None:
        super().__init__()
        self.patch_size = patch_size
        self.type = type
        self.dim = dim
        
        if type == 'conv':
            self.proj = nn.Conv2d(in_ch, dim, patch_size, patch_size, groups=patch_size*patch_size)
        else:
            self.proj = nn.ModuleList([
                nn.MaxPool2d(patch_size, patch_size),
                nn.AvgPool2d(patch_size, patch_size)
            ])
        
        self.norm = nn.LayerNorm(dim)

    def forward(self, x: Tensor) -> Tensor:
        _, _, H, W = x.shape
        if W % self.patch_size != 0:
            x = F.pad(x, (0, self.patch_size - W % self.patch_size))
        if H % self.patch_size != 0:
            x = F.pad(x, (0, 0, 0, self.patch_size - H % self.patch_size))

        if self.type == 'conv':
            x = self.proj(x)
        else:
            x = 0.5 * (self.proj[0](x) + self.proj[1](x))
        Wh, Ww = x.size(2), x.size(3)
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)
        x = x.transpose(1, 2).view(-1, self.dim, Wh, Ww)
        return x


class LawinAttn(nn.Module):
    def __init__(self, in_ch=512, head=4, patch_size=8, reduction=2) -> None:
        super().__init__()
        self.head = head

        self.position_mixing = nn.ModuleList([
            nn.Linear(patch_size * patch_size, patch_size * patch_size)
        for _ in range(self.head)])

        self.inter_channels = max(in_ch // reduction, 1)
        self.g = nn.Conv2d(in_ch, self.inter_channels, 1)
        self.theta = nn.Conv2d(in_ch, self.inter_channels, 1)
        self.phi = nn.Conv2d(in_ch, self.inter_channels, 1)
        self.conv_out = nn.Sequential(
            nn.Conv2d(self.inter_channels, in_ch, 1, bias=False),
            nn.BatchNorm2d(in_ch)
        )

    def forward(self, query: Tensor, context: Tensor) -> Tensor:
        B, C, H, W = context.shape
        context = context.reshape(B, C, -1)
        context_mlp = []

        for i, pm in enumerate(self.position_mixing):
            context_crt = context[:, (C//self.head)*i:(C//self.head)*(i+1), :]
            context_mlp.append(pm(context_crt))

        context_mlp = torch.cat(context_mlp, dim=1)
        context = context + context_mlp
        context = context.reshape(B, C, H, W)

        g_x = self.g(context).view(B, self.inter_channels, -1)
        g_x = rearrange(g_x, "b (h dim) n -> (b h) dim n", h=self.head)
        g_x = g_x.permute(0, 2, 1)

        theta_x = self.theta(query).view(B, self.inter_channels, -1)
        theta_x = rearrange(theta_x, "b (h dim) n -> (b h) dim n", h=self.head)
        theta_x = theta_x.permute(0, 2, 1)

        phi_x = self.phi(context).view(B, self.inter_channels, -1)
        phi_x = rearrange(phi_x, "b (h dim) n -> (b h) dim n", h=self.head)

        pairwise_weight = torch.matmul(theta_x, phi_x)
        pairwise_weight /= theta_x.shape[-1]**0.5
        pairwise_weight = pairwise_weight.softmax(dim=-1)

        y = torch.matmul(pairwise_weight, g_x)
        y = rearrange(y, "(b h) n dim -> b n (h dim)", h=self.head)
        y = y.permute(0, 2, 1).contiguous().reshape(B, self.inter_channels, *query.shape[-2:])

        output = query + self.conv_out(y)
        return output
        

class ConvModule(nn.Module):
    def __init__(self, c1, c2):
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, 1, bias=False)
        self.bn = nn.BatchNorm2d(c2)        # use SyncBN in original
        self.activate = nn.ReLU(True)

    def forward(self, x: Tensor) -> Tensor:
        return self.activate(self.bn(self.conv(x)))


class LawinHead(nn.Module):
    def __init__(self, in_channels: list, embed_dim=512, num_classes=19) -> None:
        super().__init__()
        for i, dim in enumerate(in_channels):
            self.add_module(f"linear_c{i+1}", MLP(dim, 48 if i == 0 else embed_dim))

        self.ratios = RATIOS
        self.lawin_8 = LawinAttn(embed_dim, 64)
        self.lawin_4 = LawinAttn(embed_dim, 16)
        self.lawin_2 = LawinAttn(embed_dim, 4)
        self.ds_8 = PatchEmbed(8, embed_dim, embed_dim)
        self.ds_4 = PatchEmbed(4, embed_dim, embed_dim)
        self.ds_2 = PatchEmbed(2, embed_dim, embed_dim)
    
        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(embed_dim, embed_dim)
        )
        self.linear_fuse = ConvModule(embed_dim*3, embed_dim)
        self.short_path = ConvModule(embed_dim, embed_dim)
        self.cat = ConvModule(embed_dim*5, embed_dim)

        self.low_level_fuse = ConvModule(embed_dim+48, embed_dim)
        self.linear_pred = nn.Conv2d(embed_dim, num_classes, 1)
        self.dropout = nn.Dropout2d(0.1)
    
    def get_lawin_att_feats(self, x: Tensor, patch_size: int, ratios: list, step: str = "") -> list:
        _, _, H, W = x.shape
        query = F.unfold(x, patch_size, stride=patch_size)
        query = rearrange(query, 'b (c ph pw) (nh nw) -> (b nh nw) c ph pw', ph=patch_size, pw=patch_size, nh=H//patch_size, nw=W//patch_size)
        outs = []

        for r in ratios:
            context = F.unfold(x, patch_size*r, stride=patch_size, padding=int((r-1)/2*patch_size))
            context = rearrange(context, "b (c ph pw) (nh nw) -> (b nh nw) c ph pw", ph=patch_size*r, pw=patch_size*r, nh=H//patch_size, nw=W//patch_size)
            context = getattr(self, f"ds_{step}{r}")(context)
            output = getattr(self, f"lawin_{step}{r}")(query, context)
            output = rearrange(output, "(b nh nw) c ph pw -> b c (nh ph) (nw pw)", ph=patch_size, pw=patch_size, nh=H//patch_size, nw=W//patch_size)
            outs.append(output)
        return outs

    def forward(self, features):
        B, _, H, W = features[1].shape
        outs = [self.linear_c2(features[1]).permute(0, 2, 1).reshape(B, -1, *features[1].shape[-2:])]

        for i, feature in enumerate(features[2:]):
            cf = eval(f"self.linear_c{i+3}")(feature).permute(0, 2, 1).reshape(B, -1, *feature.shape[-2:])
            outs.append(F.interpolate(cf, size=(H, W), mode='bilinear', align_corners=False))

        feat = self.linear_fuse(torch.cat(outs[::-1], dim=1))
        B, _, H, W = feat.shape

        ## Lawin attention spatial pyramid pooling
        feat_short = self.short_path(feat)
        feat_pool = F.interpolate(self.image_pool(feat), size=(H, W), mode='bilinear', align_corners=False)
        feat_lawin = self.get_lawin_att_feats(feat, 8, self.ratios)
        output = self.cat(torch.cat([feat_short, feat_pool, *feat_lawin], dim=1))

        ## Low-level feature enhancement
        c1 = self.linear_c1(features[0]).permute(0, 2, 1).reshape(B, -1, *features[0].shape[-2:])
        output = F.interpolate(output, size=features[0].shape[-2:], mode='bilinear', align_corners=False)
        fused = self.low_level_fuse(torch.cat([output, c1], dim=1))

        seg = self.linear_pred(self.dropout(fused))
        return seg



from torch.nn import CrossEntropyLoss, Module
import torch.nn.functional as F
import torch


class CELoss(CrossEntropyLoss):
    def __init__(self, *args, weight=None, **kwargs):
        if weight:
            weight = torch.tensor(weight)
        super().__init__(*args, weight=weight, **kwargs)


class FocalLoss(Module):
    def __init__(self, gamma: float = 0, weight=None, reduction: str = 'mean', **kwargs):
        super().__init__()
        self.weight = None
        if weight:
            self.weight = torch.tensor(weight)
        self.gamma = gamma

        if reduction == 'none':
            self.reduction = lambda x: x
        elif reduction == 'mean':
            self.reduction = torch.mean
        elif reduction == 'sum':
            self.reduction = torch.sum
        else:
            raise NotImplementedError(f"Invalid reduction mode: {self.reduction}")

    def __call__(self, x, target, **kwargs):
        ce_loss = F.cross_entropy(x, target.float(), reduction='none', **kwargs)
        pt = torch.exp(-ce_loss)
        if self.weight is not None:
            wtarget = self.weight[(...,) + (None, ) * (len(target.shape) - 1)]\
                          .moveaxis(0, 1).to(target.device) \
                      * target
            wtarget = wtarget[wtarget != 0].reshape(pt.shape)
            focal_loss = torch.pow((1 - pt), self.gamma) * wtarget * ce_loss
        else:
            focal_loss = torch.pow((1 - pt), self.gamma) * ce_loss

        return self.reduction(focal_loss)


LOSSES = {
    'cross_entropy': CELoss,
    'focal': FocalLoss
} 



from super_gradients.training import dataloaders

from ezdl.data import DatasetInterface


class Cifar10(DatasetInterface):
    def __init__(self, dataset_params):
        super().__init__(dataset_params)
        self.train_loader = dataloaders.get("cifar10_train", dataset_params=dataset_params['trainset'],
                                            dataloader_params=dataset_params['trainloader'])
        self.test_loader = dataloaders.get("cifar10_val", dataset_params=dataset_params['testset'],
                                           dataloader_params=dataset_params['testloader'])
        self.val_loader = dataloaders.get("cifar10_val", dataset_params=dataset_params['testset'],
                                           dataloader_params=dataset_params['testloader'])


import importlib
from typing import Mapping, Dict

import pandas as pd
from super_gradients.common import MultiGPUMode

from ezdl.callbacks import SegmentationVisualizationCallback, callback_factory
from ezdl.utils.utilities import get_module_class_from_path
from piptools.scripts.sync import _get_installed_distributions

from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.common.sg_loggers import SG_LOGGERS, BaseSGLogger
from super_gradients.common.sg_loggers.abstract_sg_logger import AbstractSGLogger
from super_gradients.training import StrictLoad, Trainer

from super_gradients.training.params import TrainingParams
from super_gradients.training.utils.callbacks import Phase, PhaseContext, CallbackHandler
from super_gradients.training import utils as core_utils
from super_gradients.training import models

import torch
from super_gradients.training.utils.checkpoint_utils import load_checkpoint_to_model
from tqdm import tqdm

# from ezdl.callbacks import SegmentationVisualizationCallback
from ezdl.models import MODELS as MODELS_DICT
from ezdl.logger.basesg_logger import BaseSGLogger as BaseLogger
from ezdl.logger import LOGGERS

logger = get_logger(__name__)


class SegmentationTrainer(Trainer):
    def __init__(self, ckpt_root_dir=None, model_checkpoints_location: str = 'local', **kwargs):
        self.run_id = None
        self.train_initialized = False
        self.model_checkpoints_location = model_checkpoints_location
        super().__init__(ckpt_root_dir=ckpt_root_dir, **kwargs)

    def init_dataset(self, dataset_interface, dataset_params):
        if isinstance(dataset_interface, str):
            try:
                module, dataset_interface_cls = get_module_class_from_path(dataset_interface)
                dataset_module = importlib.import_module(module)
                dataset_interface_cls = getattr(dataset_module, dataset_interface_cls)
            except AttributeError as e:
                raise AttributeError("No interface found!") from e
            dataset_interface = dataset_interface_cls(dataset_params)
        elif isinstance(dataset_interface, type):
            pass
        else:
            raise ValueError("Dataset interface should be str or class")
        data_loader_num_workers = dataset_params.get('num_workers') or 0
        self.connect_dataset_interface(dataset_interface, data_loader_num_workers)
        return self.dataset_interface

    def connect_dataset_interface(self, dataset_interface, data_loader_num_workers):
        self.dataset_interface = dataset_interface
        self.train_loader, self.valid_loader, self.test_loader, self.classes = \
            self.dataset_interface.get_data_loaders(batch_size_factor=self.num_devices,
                                                    num_workers=data_loader_num_workers,
                                                    distributed_sampler=self.multi_gpu == MultiGPUMode.DISTRIBUTED_DATA_PARALLEL)

        self.dataset_params = self.dataset_interface.get_dataset_params()

    def init_model(self, params: Mapping, resume: bool, checkpoint_path: str = None):
        # init model
        model_params = params['model']
        input_channels = len(params['dataset']['channels'])
        output_channels = params['dataset']['num_classes']
        arch_params = {
            'input_channels': input_channels,
            'output_channels': output_channels,
            'in_channels': input_channels,
            'out_channels': output_channels,
            'num_classes': output_channels,
            **model_params['params']
        }
        try:
            module, model_cls = get_module_class_from_path(model_params['name'])
            model_module = importlib.import_module(module)
            model_cls = getattr(model_module, model_cls)
            model = model_cls(arch_params)
        except (AttributeError, ValueError):
            if model_params['name'] in MODELS_DICT.keys():
                model = MODELS_DICT[model_params['name']](arch_params)
            else:
                model = models.get(model_name=model_params['name'], arch_params=arch_params)

        if 'num_classes' not in arch_params.keys():
            if self.classes is None and self.dataset_interface is None:
                raise Exception('Error', 'Number of classes not defined in arch params and dataset is not defined')
            else:
                arch_params['num_classes'] = len(self.classes)

        self.arch_params = core_utils.HpmStruct(**arch_params)
        self.net = model

        self._net_to_device()
        # SET THE FLAG FOR DIFFERENT PARAMETER GROUP OPTIMIZER UPDATE
        self.update_param_groups = hasattr(self.net.module, 'update_param_groups')

        if resume and checkpoint_path is not None:
            self.checkpoint = load_checkpoint_to_model(ckpt_local_path=checkpoint_path,
                                                       load_backbone=False,
                                                       net=self.net,
                                                       strict=StrictLoad.ON.value,
                                                       load_weights_only=self.load_weights_only)
            self.load_checkpoint = True

            if 'ema_net' in self.checkpoint.keys():
                logger.warning(
                    "[WARNING] Main network has been loaded from checkpoint but EMA network exists as well. It "
                    " will only be loaded during validation when training with ema=True. ")

            # UPDATE TRAINING PARAMS IF THEY EXIST & WE ARE NOT LOADING AN EXTERNAL MODEL's WEIGHTS
            self.best_metric = self.checkpoint['acc'] if 'acc' in self.checkpoint.keys() else -1
            self.start_epoch = self.checkpoint['epoch'] if 'epoch' in self.checkpoint.keys() else 0

    def train(self, training_params: dict = dict()):
        super().train(model=self.net,
                      training_params=training_params,
                      train_loader=self.train_loader,
                      valid_loader=self.valid_loader)
        if self.train_loader.num_workers > 0 and self.train_loader._iterator is not None:
            self.train_loader._iterator._shutdown_workers()
        if self.valid_loader.num_workers > 0 and self.valid_loader._iterator is not None:
            self.valid_loader._iterator._shutdown_workers()
        # Restore best parameters
        self.checkpoint = load_checkpoint_to_model(ckpt_local_path=self.checkpoints_dir_path + '/ckpt_best.pth',
                                                   load_backbone=False,
                                                   net=self.net,
                                                   strict=StrictLoad.ON.value,
                                                   load_weights_only=True)

    def test(self,  # noqa: C901
             test_loader: torch.utils.data.DataLoader = None,
             loss: torch.nn.modules.loss._Loss = None,
             silent_mode: bool = False,
             test_metrics: Mapping = None,
             loss_logging_items_names=None, metrics_progress_verbose=False, test_phase_callbacks={},
             use_ema_net=True) -> dict:
        """
        Evaluates the model on given dataloader and metrics.

        :param test_loader: dataloader to perform test on.
        :param test_metrics: dict name: Metric for evaluation.
        :param silent_mode: (bool) controls verbosity
        :param metrics_progress_verbose: (bool) controls the verbosity of metrics progress (default=False). Slows down the program.
        :param use_ema_net (bool) whether to perform test on self.ema_model.ema (when self.ema_model.ema exists,
            otherwise self.net will be tested) (default=True)
        :return: results tuple (tuple) containing the loss items and metric values.

        All of the above args will override SgModel's corresponding attribute when not equal to None. Then evaluation
         is ran on self.test_loader with self.test_metrics.
        """
        test_loader = test_loader or self.test_loader
        loss = loss or self.training_params.loss
        if loss is not None:
            loss.to(self.device)
        test_phase_callbacks = [callback_factory(name, params, seg_trainer=self, dataset=self.dataset_interface, loader=test_loader)
                                for name, params in test_phase_callbacks.items()]
        metrics_values = super().test(model=self.net,
                                      test_loader=test_loader,
                                      loss=loss,
                                      silent_mode=silent_mode,
                                      test_metrics_list=list(test_metrics.values()),
                                      loss_logging_items_names=loss_logging_items_names,
                                      metrics_progress_verbose=metrics_progress_verbose,
                                      test_phase_callbacks=test_phase_callbacks,
                                      use_ema_net=use_ema_net)

        metric_names = test_metrics.keys()

        if self.test_loader.num_workers > 0:
            self.test_loader._iterator._shutdown_workers()
        metrics = {'test_loss': metrics_values[0], **dict(zip(metric_names, metrics_values[1:]))}
        if 'conf_mat' in metrics.keys():
            metrics.pop('conf_mat')
            cf = test_metrics['conf_mat'].get_cf()
            logger.info(f'Confusion matrix:\n{cf}')
            self.sg_logger.add_table('confusion_matrix', cf.cpu(),
                                     columns=list(self.dataset_interface.testset.CLASS_LABELS.values()),
                                     rows=list(self.dataset_interface.testset.CLASS_LABELS.values())
                                     )
        self.sg_logger.add_summary(metrics)
        if 'auc' in metrics.keys():
            logger.info('Computing ROC curve...')
            roc = test_metrics['auc'].get_roc()
            fpr_tpr = [(roc[0][i], roc[1][i]) for i in range(len(roc))]
            skip = [x[0].shape.numel() // 1000 for x in fpr_tpr]
            fpr_tpr = [(fpr[::sk], tpr[::sk]) for (fpr, tpr), sk in zip(fpr_tpr, skip)]
            fprs, tprs = zip(*fpr_tpr)
            fprs = torch.cat(fprs)
            tprs = torch.cat(tprs)
            classes = list(self.dataset_interface.testset.CLASS_LABELS.values())
            cls = [[classes[i]] * len(fpr)
                   for i, (fpr, tpr) in enumerate(fpr_tpr)]
            cls = [item for sublist in cls for item in sublist]
            df = pd.DataFrame({'class': cls, 'fpr': fprs, 'tpr': tprs})
            name = "ROCAUC"
            self.sg_logger.add_plot(name, df, "fpr", "tpr", classes_marker="class")
            logger.info('ROC curve computed.')
        return metrics

    def _init_monitored_items(self):
        if self.metric_to_watch == 'loss':
            if len(self.loss_logging_items_names) == 1:
                self.metric_to_watch = self.loss_logging_items_names[0]
            else:
                raise ValueError(f"Specify which loss {self.loss_logging_items_names} to watch")
        return super()._init_monitored_items()

    def _initialize_sg_logger_objects(self, additional_configs_to_log: Dict = None):
        if not self.train_initialized:
            self.train_initialized = True
            # OVERRIDE SOME PARAMETERS TO MAKE SURE THEY MATCH THE TRAINING PARAMETERS
            general_sg_logger_params = {  # 'experiment_name': self.experiment_name,
                'experiment_name': '',
                'group': self.experiment_name,
                'storage_location': self.model_checkpoints_location,
                'resumed': self.load_checkpoint,
                'training_params': self.training_params,
                'checkpoints_dir_path': self.checkpoints_dir_path,
                'run_id': self.run_id
            }
            sg_logger = core_utils.get_param(self.training_params, 'sg_logger')

            if sg_logger is None:
                raise RuntimeError('logger must be defined in experiment params (see default_training_params)')

            if isinstance(sg_logger, AbstractSGLogger):
                self.sg_logger = sg_logger
            elif isinstance(sg_logger, str):
                sg_logger_params = {**general_sg_logger_params,
                                    **core_utils.get_param(self.training_params, 'sg_logger_params', {})}
                if sg_logger in SG_LOGGERS:
                    self.sg_logger = SG_LOGGERS[sg_logger](**sg_logger_params)
                elif sg_logger in LOGGERS:
                    self.sg_logger = LOGGERS[sg_logger](**sg_logger_params)
                else:
                    raise RuntimeError('sg_logger not defined in SG_LOGGERS of SuperGradients neither in LOGGERS of '
                                       'EzDL')
            else:
                raise RuntimeError('sg_logger can be either an sg_logger name (str) or a subcalss of AbstractSGLogger')

            if not (isinstance(self.sg_logger, BaseSGLogger) or isinstance(self.sg_logger, BaseLogger)):
                logger.warning(
                    "WARNING! Using a user-defined sg_logger: files will not be automatically written to disk!\n"
                    "Please make sure the provided sg_logger writes to disk or compose your sg_logger to BaseSGLogger")

            # IN CASE SG_LOGGER UPDATED THE DIR PATH
            self.checkpoints_dir_path = self.sg_logger.local_dir()
            self.training_params.override(sg_logger=sg_logger)
            additional_log_items = {'num_devices': self.num_devices,
                                    'multi_gpu': str(self.multi_gpu),
                                    'device_type': torch.cuda.get_device_name(
                                        0) if torch.cuda.is_available() else 'cpu'}

            # ADD INSTALLED PACKAGE LIST + THEIR VERSIONS
            if self.training_params.log_installed_packages:
                pkg_list = list(map(lambda pkg: str(pkg), _get_installed_distributions()))
                additional_log_items['installed_packages'] = pkg_list

            self.sg_logger.add_config("additional_log_items", additional_log_items)
            self.sg_logger.flush()

    def init_loggers(self,
                     in_params: Mapping = None,
                     train_params: Mapping = None,
                     init_sg_loggers: bool = True,
                     run_id=None) -> None:

        self.run_id = run_id
        if self.training_params is None:
            self.training_params = TrainingParams()
        self.training_params.override(**train_params)
        if init_sg_loggers:
            self._initialize_sg_logger_objects()
        if self.phase_callbacks is None:
            self.phase_callbacks = []
        self.phase_callback_handler = CallbackHandler(self.phase_callbacks)
        self.sg_logger.add_config(config=in_params)

    def run(self, data_loader: torch.utils.data.DataLoader, callbacks=None, silent_mode: bool = False):
        """
        Runs the model on given dataloader.

        :param data_loader: dataloader to perform run on

        """

        # THE DISABLE FLAG CONTROLS WHETHER THE PROGRESS BAR IS SILENT OR PRINTS THE LOGS
        if callbacks is None:
            callbacks = []
        progress_bar_data_loader = tqdm(data_loader, bar_format="{l_bar}{bar:10}{r_bar}", dynamic_ncols=True,
                                        disable=silent_mode)
        context = PhaseContext(criterion=self.criterion,
                               device=self.device,
                               sg_logger=self.sg_logger)

        self.phase_callbacks.extend(callbacks)

        if not silent_mode:
            # PRINT TITLES
            pbar_start_msg = f"Running model on {len(data_loader)} batches"
            progress_bar_data_loader.set_description(pbar_start_msg)

        with torch.no_grad():
            for batch_idx, batch_items in enumerate(progress_bar_data_loader):
                batch_items = core_utils.tensor_container_to_device(batch_items, self.device, non_blocking=True)

                additional_batch_items = {}
                targets = None
                if hasattr(batch_items, '__len__'):
                    if len(batch_items) == 2:
                        inputs, targets = batch_items
                    elif len(batch_items) == 3:
                        inputs, targets, additional_batch_items = batch_items
                    else:
                        raise ValueError(f"Expected 1, 2 or 3 items in batch_items, got {len(batch_items)}")
                else:
                    inputs = batch_items

                output = self.net(inputs)

                context.update_context(batch_idx=batch_idx,
                                       inputs=inputs,
                                       preds=output,
                                       target=targets,
                                       **additional_batch_items)

                # TRIGGER PHASE CALLBACKS
                self.phase_callback_handler(Phase.POST_TRAINING, context)
